"""Register available envs and create index of train/test mappings."""

import collections
import functools
import re

import cv2
import gym
from gym.spaces import Box, Dict
from gym.wrappers import ResizeObservation
import numpy as np

from magical.benchmarks.cluster import ClusterColourEnv, ClusterShapeEnv
from magical.benchmarks.find_dupe import FindDupeEnv
from magical.benchmarks.fix_colour import FixColourEnv
from magical.benchmarks.make_line import MakeLineEnv
from magical.benchmarks.match_regions import MatchRegionsEnv
from magical.benchmarks.move_to_corner import MoveToCornerEnv
from magical.benchmarks.move_to_region import MoveToRegionEnv

__all__ = [
    'DEMO_ENVS_TO_TEST_ENVS_MAP',
    'MoveToCornerEnv',
    'MatchRegionsEnv',
    'ClusterColourEnv',
    'ClusterShapeEnv',
    'FindDupeEnv',
    'MakeLineEnv',
    'register_envs',
    'EnvName',
]

DEFAULT_RES = (384, 384)


def _gym_tree_map(f, *structures):
    """Apply a function f to the given structures. If the structures are
    dictionaries or Gym dictionary observation spaces, then it recursively
    applies the function to their values and returns a new dict (or new Dict
    observation space), as appropriate."""
    s0 = structures[0]
    if isinstance(s0, Dict):
        return Dict(
            collections.OrderedDict([
                (k,
                 _gym_tree_map(f, *(struct.spaces[k]
                                    for struct in structures)))
                for k in s0.spaces.keys()
            ]))
    elif isinstance(s0, (dict, collections.OrderedDict)):
        return type(s0)((k, _gym_tree_map(f, *(s[k] for s in structures)))
                        for k in s0.keys())
    return f(*structures)


class EagerDictFrameStack(gym.Wrapper):
    """Version of Gym's frame stack wrapper that is (1) totally eager, (2)
    stacks along channels axis instead of a separate leading axis, and (3)
    supports nested Dict observation spaces with Box values at the leaves."""
    def __init__(self, env, depth):
        super().__init__(env)
        self.depth = depth
        self.frames = collections.deque(maxlen=depth)

        def box_map(box):
            low = np.repeat(box.low, depth, axis=-1)
            high = np.repeat(box.high, depth, axis=-1)
            return Box(low=low, high=high, dtype=box.dtype)

        self.observation_space = _gym_tree_map(box_map, env.observation_space)

    def _get_observation(self):
        assert len(self.frames) == self.depth, \
            (len(self.frames), self.depth)
        return _gym_tree_map(lambda *frames: np.concatenate(frames, axis=-1),
                             *self.frames)

    def step(self, action):
        observation, reward, done, info = self.env.step(action)
        self.frames.append(observation)
        return self._get_observation(), reward, done, info

    def reset(self, **kwargs):
        observation = self.env.reset(**kwargs)
        for _ in range(self.depth):
            self.frames.append(observation)
        return self._get_observation()


class FlattenFrameStack(gym.Wrapper):
    """Like EagerFrameStack, except it stacks all images into one big frame.
    Supports variable lookback for each key."""
    def __init__(self, env, depth_by_key):
        super().__init__(env)
        self.depth_by_key = depth_by_key
        self.frames_by_key = collections.OrderedDict([
            (k, collections.deque(maxlen=k_depth))
            for k, k_depth in depth_by_key.items()
        ])

        orig_space = None

        def box_map(box):
            nonlocal orig_space
            if orig_space is None:
                orig_space = box
            else:
                assert np.all(box.low == orig_space.low)
                assert np.all(box.high == orig_space.high)
                assert box.dtype == orig_space.dtype
            return box  # this is never used

        # figure out what the space is inside the dict
        assert isinstance(env.observation_space, Dict)
        _gym_tree_map(box_map, env.observation_space)
        self.depth_sum = sum(self.depth_by_key.values())
        new_low = np.repeat(orig_space.low, self.depth_sum, axis=-1)
        new_high = np.repeat(orig_space.high, self.depth_sum, axis=-1)
        self.observation_space = Box(low=new_low,
                                     high=new_high,
                                     dtype=orig_space.dtype)

    def _get_observation(self):
        # assume depth 1 dict
        all_frames = []
        for frames in self.frames_by_key.values():
            all_frames.extend(frames)
        len(all_frames) == self.depth_sum
        stacked_image = np.concatenate(all_frames, axis=-1)
        return stacked_image

    def step(self, action):
        observation, reward, done, info = self.env.step(action)
        for key, frame in observation.items():
            self.frames_by_key[key].append(frame)
        return self._get_observation(), reward, done, info

    def reset(self, **kwargs):
        observation = self.env.reset(**kwargs)
        for key, frame in observation.items():
            depth = self.depth_by_key[key]
            for _ in range(depth):
                self.frames_by_key[key].append(frame)
        return self._get_observation()


class ResizeDictObservation(gym.ObservationWrapper):
    """Version of Gym's ResizeObservation wrapper that supports dict
    observations."""
    def __init__(self, env, res):
        super().__init__(env)

        if isinstance(res, int):
            res = (res, res)
        assert all(x > 0 for x in res), res
        self.res = tuple(res)

        def space_mapper(box):
            # copy channels from the box, but replace height/width (assumed to
            # be leading two dims)
            new_shape = self.res + box.shape[2:]
            return Box(low=0, high=255, shape=new_shape, dtype=np.uint8)

        self.observation_space = _gym_tree_map(space_mapper,
                                               env.observation_space)

    def observation(self, observation):
        # use exactly the same resizing method as ResizeObservation in Gym
        def obs_mapper(box_obs):
            box_obs = cv2.resize(box_obs,
                                 self.res[::-1],
                                 interpolation=cv2.INTER_AREA)
            if box_obs.ndim == 2:
                box_obs = np.expand_dims(box_obs, -1)
            return box_obs

        return _gym_tree_map(obs_mapper, observation)


class ChannelsFirst(gym.ObservationWrapper):
    """Moves last axis around to first position. Useful for going from
    channels-last to channels-first representation."""
    def __init__(self, env):
        super().__init__(env)

        def space_mapper(box):
            assert isinstance(box, gym.spaces.Box), box
            return gym.spaces.Box(low=self._rotate(box.low),
                                  high=self._rotate(box.high),
                                  dtype=box.dtype)

        self.observation_space = _gym_tree_map(space_mapper,
                                               env.observation_space)

    @staticmethod
    def _rotate(array):
        return np.moveaxis(array, -1, 0)

    def observation(self, observation):
        return _gym_tree_map(self._rotate, observation)


def lores_stack_entry_point(env_cls, small_res, frames=4):
    def make_lores_stack(**kwargs):
        base_env = env_cls(**kwargs)
        resize_env = ResizeDictObservation(base_env, small_res)
        stack_env = EagerDictFrameStack(resize_env, frames)
        return stack_env

    return make_lores_stack


def lores_ea_entry_point(env_cls,
                         small_res,
                         allo_frames=1,
                         ego_frames=3,
                         channels_first=False):
    """For stacking ego/allo frames together."""
    def make_lores_ea(**kwargs):
        base_env = env_cls(**kwargs)
        stack_env = FlattenFrameStack(
            base_env,
            collections.OrderedDict([
                ('allo', allo_frames),
                ('ego', ego_frames),
            ]))
        resize_env = ResizeObservation(stack_env, small_res)
        if channels_first:
            return ChannelsFirst(resize_env)
        return resize_env

    return make_lores_ea


DEFAULT_PREPROC_ENTRY_POINT_WRAPPERS = collections.OrderedDict([
    # stacking latest allo view with three most recent ego views (output is
    # just a stacked array, not a 3D array)
    ('LoRes3EA',
     functools.partial(lores_ea_entry_point,
                       small_res=(96, 96),
                       allo_frames=1,
                       ego_frames=3)),
    # stacking egocentric views from four ego/allo dicts & stacking them
    # together
    ('LoRes4E',
     functools.partial(lores_ea_entry_point,
                       small_res=(96, 96),
                       allo_frames=0,
                       ego_frames=4)),
    # stacking four allocentric views instead
    ('LoRes4A',
     functools.partial(lores_ea_entry_point,
                       small_res=(96, 96),
                       allo_frames=4,
                       ego_frames=0)),
    # stacking values of four dicts together to produce a single new dict
    ('LoResStack',
     functools.partial(lores_stack_entry_point, small_res=(96, 96), frames=4)),
    # stacking egocentric views from four ego/allo dicts & stacking them
    # together
    ('LoResCHW4E',
     functools.partial(lores_ea_entry_point,
                       small_res=(96, 96),
                       allo_frames=0,
                       ego_frames=4,
                       channels_first=True)),
])
_ENV_NAME_RE = re.compile(
    r'^(?P<name_prefix>[^-]+)(?P<demo_test_spec>-(Demo|Test[^-]*))'
    r'(?P<env_name_suffix>(-[^-]+)*)(?P<version_suffix>-v\d+)$')
_REGISTERED = False
# this will be filled in later
DEMO_ENVS_TO_TEST_ENVS_MAP = collections.OrderedDict()
AVAILABLE_PREPROCESSORS = [key for key in DEFAULT_PREPROC_ENTRY_POINT_WRAPPERS]


class EnvName:
    """Convenience class for parsing environment names. All environment names
    look like this (per _ENV_NAME_RE):

        <name_prefix>-<demo_test_spec>[-<suffix>]-<version_suffix>

    Where:
        - name_prefix identifies the environment class (e.g. MoveToCorner).
        - demo_test_spec is either 'demo' for the demonstration environment, or
          'Test<description>' for test environments, where the description is
          something like 'Pose' or 'Colour' or 'All', depending on what aspects
          of the environment are randomised.
        - suffix is usually for indicating a type of postprocessing (e.g.
          -LoResStack for stacked, scaled-down frames). Not always present.
        - version_suffix is -v0, -v1, etc. etc."""
    def __init__(self, env_name):
        match = _ENV_NAME_RE.match(env_name)
        if match is None:
            raise ValueError(
                "env name '{env_name}' does not match _ENV_NAME_RE spec")
        groups = match.groupdict()
        self.env_name = env_name
        self.name_prefix = groups['name_prefix']
        self.demo_test_spec = groups['demo_test_spec']
        self.env_name_suffix = groups['env_name_suffix']
        self.version_suffix = groups['version_suffix']
        self.demo_env_name = self.name_prefix + '-Demo' \
            + self.env_name_suffix + self.version_suffix
        self.is_test = self.demo_test_spec.startswith('-Test')
        if not self.is_test:
            assert self.demo_env_name == self.env_name, \
                (self.demo_env_name, self.env_name)


def register_envs():
    """Register all default environments for this benchmark suite."""
    global _REGISTERED, DEMO_ENVS_TO_TEST_ENVS_MAP
    if _REGISTERED:
        return False
    _REGISTERED = True

    common_kwargs = dict(res_hw=DEFAULT_RES,
                         fps=8,
                         phys_steps=10,
                         phys_iter=10)

    # remember 100 frames is ~12.5s at 8fps
    mtc_ep_len = 80
    move_to_corner_variants = [
        (MoveToCornerEnv, mtc_ep_len, '-Demo', {
            'rand_shape_colour': False,
            'rand_shape_type': False,
            'rand_poses': False,
            'rand_dynamics': False,
        }),
        (MoveToCornerEnv, mtc_ep_len, '-TestColour', {
            'rand_shape_colour': True,
            'rand_shape_type': False,
            'rand_poses': False,
            'rand_dynamics': False,
        }),
        (MoveToCornerEnv, mtc_ep_len, '-TestShape', {
            'rand_shape_colour': False,
            'rand_shape_type': True,
            'rand_poses': False,
            'rand_dynamics': False,
        }),
        (MoveToCornerEnv, mtc_ep_len, '-TestJitter', {
            'rand_shape_colour': False,
            'rand_shape_type': False,
            'rand_poses': True,
            'rand_dynamics': False,
        }),
        (MoveToCornerEnv, mtc_ep_len, '-TestDynamics', {
            'rand_shape_colour': False,
            'rand_shape_type': False,
            'rand_poses': False,
            'rand_dynamics': True,
        }),
        (MoveToCornerEnv, mtc_ep_len, '-TestAll', {
            'rand_shape_colour': True,
            'rand_shape_type': True,
            'rand_poses': True,
            'rand_dynamics': True,
        }),
    ]

    mtr_ep_len = 40
    move_to_region_variants = [
        (MoveToRegionEnv, mtr_ep_len, '-Demo', {
            'rand_poses_minor': False,
            'rand_poses_full': False,
            'rand_goal_colour': False,
            'rand_dynamics': False,
        }),
        (MoveToRegionEnv, mtr_ep_len, '-TestJitter', {
            'rand_poses_minor': True,
            'rand_poses_full': False,
            'rand_goal_colour': False,
            'rand_dynamics': False,
        }),
        (MoveToRegionEnv, mtr_ep_len, '-TestColour', {
            'rand_poses_minor': False,
            'rand_poses_full': False,
            'rand_goal_colour': True,
            'rand_dynamics': False,
        }),
        (MoveToRegionEnv, mtr_ep_len, '-TestLayout', {
            'rand_poses_minor': False,
            'rand_poses_full': True,
            'rand_goal_colour': False,
            'rand_dynamics': False,
        }),
        (MoveToRegionEnv, mtr_ep_len, '-TestDynamics', {
            'rand_poses_minor': False,
            'rand_poses_full': False,
            'rand_goal_colour': False,
            'rand_dynamics': True,
        }),
        (MoveToRegionEnv, mtr_ep_len, '-TestAll', {  # to stop yapf
            'rand_poses_minor': False,
            # rand_poses_full subsumes rand_poses_minor
            'rand_poses_full': True,
            'rand_goal_colour': True,
            'rand_dynamics': True,
        }),
        # egocentric view shift (probably a bad idea to add this as a test env,
        # rather than a postprocessor or something)
        # (MoveToRegionEnv, mtr_ep_len, '-TestEgo', {
        #     'rand_poses_minor': False,
        #     'rand_poses_full': False,
        #     'rand_goal_colour': False,
        #     'rand_dynamics': False,
        #     'egocentric': True,
        # }),
    ]

    mr_ep_len = 120
    match_regions_variants = [
        (MatchRegionsEnv, mr_ep_len, '-Demo', {
            'rand_target_colour': False,
            'rand_shape_type': False,
            'rand_shape_count': False,
            'rand_layout_minor': False,
            'rand_layout_full': False,
            'rand_dynamics': False,
        }),
        (MatchRegionsEnv, mr_ep_len, '-TestJitter', {
            'rand_target_colour': False,
            'rand_shape_type': False,
            'rand_shape_count': False,
            'rand_layout_minor': True,
            'rand_layout_full': False,
            'rand_dynamics': False,
        }),
        (MatchRegionsEnv, mr_ep_len, '-TestColour', {
            'rand_target_colour': True,
            'rand_shape_type': False,
            'rand_shape_count': False,
            'rand_layout_minor': False,
            'rand_layout_full': False,
            'rand_dynamics': False,
        }),
        (MatchRegionsEnv, mr_ep_len, '-TestShape', {
            'rand_target_colour': False,
            'rand_shape_type': True,
            'rand_shape_count': False,
            'rand_layout_minor': False,
            'rand_layout_full': False,
            'rand_dynamics': False,
        }),
        (MatchRegionsEnv, mr_ep_len, '-TestLayout', {
            'rand_target_colour': False,
            'rand_shape_type': False,
            'rand_shape_count': False,
            'rand_layout_minor': False,
            'rand_layout_full': True,
            'rand_dynamics': False,
        }),
        # test everything except dynamics
        (MatchRegionsEnv, mr_ep_len, '-TestCountPlus', {
            'rand_target_colour': True,
            'rand_shape_type': True,
            'rand_shape_count': True,
            'rand_layout_minor': False,
            'rand_layout_full': True,
            'rand_dynamics': False,
        }),
        (MatchRegionsEnv, mr_ep_len, '-TestDynamics', {
            'rand_target_colour': False,
            'rand_shape_type': False,
            'rand_shape_count': False,
            'rand_layout_minor': False,
            'rand_layout_full': False,
            'rand_dynamics': True,
        }),
        (MatchRegionsEnv, mr_ep_len, '-TestAll', {
            'rand_target_colour': True,
            'rand_shape_type': True,
            'rand_shape_count': True,
            'rand_layout_minor': False,
            'rand_layout_full': True,
            'rand_dynamics': True,
        }),
    ]

    ml_ep_len = 180
    make_line_variants = [
        (MakeLineEnv, ml_ep_len, '-Demo', {
            'rand_colours': False,
            'rand_shapes': False,
            'rand_count': False,
            'rand_layout_minor': False,
            'rand_layout_full': False,
            'rand_dynamics': False,
        }),
        (MakeLineEnv, ml_ep_len, '-TestJitter', {
            'rand_colours': False,
            'rand_shapes': False,
            'rand_count': False,
            'rand_layout_minor': True,
            'rand_layout_full': False,
            'rand_dynamics': False,
        }),
        (MakeLineEnv, ml_ep_len, '-TestColour', {
            'rand_colours': True,
            'rand_shapes': False,
            'rand_count': False,
            'rand_layout_minor': False,
            'rand_layout_full': False,
            'rand_dynamics': False,
        }),
        (MakeLineEnv, ml_ep_len, '-TestShape', {
            'rand_colours': False,
            'rand_shapes': True,
            'rand_count': False,
            'rand_layout_minor': False,
            'rand_layout_full': False,
            'rand_dynamics': False,
        }),
        (MakeLineEnv, ml_ep_len, '-TestLayout', {
            'rand_colours': False,
            'rand_shapes': False,
            'rand_count': False,
            'rand_layout_minor': False,
            'rand_layout_full': True,
            'rand_dynamics': False,
        }),
        # test everything except dynamics
        (MakeLineEnv, ml_ep_len, '-TestCountPlus', {
            'rand_colours': True,
            'rand_shapes': True,
            'rand_count': True,
            'rand_layout_minor': False,
            'rand_layout_full': True,
            'rand_dynamics': False,
        }),
        (MakeLineEnv, ml_ep_len, '-TestDynamics', {
            'rand_colours': False,
            'rand_shapes': False,
            'rand_count': False,
            'rand_layout_minor': False,
            'rand_layout_full': False,
            'rand_dynamics': True,
        }),
        (MakeLineEnv, ml_ep_len, '-TestAll', {
            'rand_colours': True,
            'rand_shapes': True,
            'rand_count': True,
            'rand_layout_minor': False,
            'rand_layout_full': True,
            'rand_dynamics': True,
        }),
    ]

    fd_ep_len = 100
    find_dupe_variants = [
        (FindDupeEnv, fd_ep_len, '-Demo', {
            'rand_colours': False,
            'rand_shapes': False,
            'rand_count': False,
            'rand_layout_minor': False,
            'rand_layout_full': False,
            'rand_dynamics': False,
        }),
        (FindDupeEnv, fd_ep_len, '-TestJitter', {
            'rand_colours': False,
            'rand_shapes': False,
            'rand_count': False,
            'rand_layout_minor': True,
            'rand_layout_full': False,
            'rand_dynamics': False,
        }),
        (FindDupeEnv, fd_ep_len, '-TestColour', {
            'rand_colours': True,
            'rand_shapes': False,
            'rand_count': False,
            'rand_layout_minor': False,
            'rand_layout_full': False,
            'rand_dynamics': False,
        }),
        (FindDupeEnv, fd_ep_len, '-TestShape', {
            'rand_colours': False,
            'rand_shapes': True,
            'rand_count': False,
            'rand_layout_minor': False,
            'rand_layout_full': False,
            'rand_dynamics': False,
        }),
        (FindDupeEnv, fd_ep_len, '-TestLayout', {
            'rand_colours': False,
            'rand_shapes': False,
            'rand_count': False,
            'rand_layout_minor': False,
            'rand_layout_full': True,
            'rand_dynamics': False,
        }),
        (FindDupeEnv, fd_ep_len, '-TestCountPlus', {
            'rand_colours': True,
            'rand_shapes': True,
            'rand_count': True,
            'rand_layout_minor': False,
            'rand_layout_full': True,
            'rand_dynamics': False,
        }),
        (FindDupeEnv, fd_ep_len, '-TestDynamics', {
            'rand_colours': False,
            'rand_shapes': False,
            'rand_count': False,
            'rand_layout_minor': False,
            'rand_layout_full': False,
            'rand_dynamics': True,
        }),
        (FindDupeEnv, fd_ep_len, '-TestAll', {
            'rand_colours': True,
            'rand_shapes': True,
            'rand_count': True,
            'rand_layout_minor': False,
            'rand_layout_full': True,
            'rand_dynamics': True,
        }),
    ]

    fc_ep_len = 60
    fix_colour_variants = [
        (FixColourEnv, fc_ep_len, '-Demo', {
            'rand_colours': False,
            'rand_shapes': False,
            'rand_count': False,
            'rand_layout_minor': False,
            'rand_layout_full': False,
            'rand_dynamics': False,
        }),
        (FixColourEnv, fc_ep_len, '-TestJitter', {
            'rand_colours': False,
            'rand_shapes': False,
            'rand_count': False,
            'rand_layout_minor': True,
            'rand_layout_full': False,
            'rand_dynamics': False,
        }),
        (FixColourEnv, fc_ep_len, '-TestColour', {
            'rand_colours': True,
            'rand_shapes': False,
            'rand_count': False,
            'rand_layout_minor': False,
            'rand_layout_full': False,
            'rand_dynamics': False,
        }),
        (FixColourEnv, fc_ep_len, '-TestShape', {
            'rand_colours': False,
            'rand_shapes': True,
            'rand_count': False,
            'rand_layout_minor': False,
            'rand_layout_full': False,
            'rand_dynamics': False,
        }),
        (FixColourEnv, fc_ep_len, '-TestLayout', {
            'rand_colours': False,
            'rand_shapes': False,
            'rand_count': False,
            'rand_layout_minor': False,
            'rand_layout_full': True,
            'rand_dynamics': False,
        }),
        (FixColourEnv, fc_ep_len, '-TestCountPlus', {
            'rand_colours': True,
            'rand_shapes': True,
            'rand_count': True,
            'rand_layout_minor': False,
            'rand_layout_full': True,
            'rand_dynamics': False,
        }),
        (FixColourEnv, fc_ep_len, '-TestDynamics', {
            'rand_colours': False,
            'rand_shapes': False,
            'rand_count': False,
            'rand_layout_minor': False,
            'rand_layout_full': False,
            'rand_dynamics': True,
        }),
        (FixColourEnv, fc_ep_len, '-TestAll', {
            'rand_colours': True,
            'rand_shapes': True,
            'rand_count': True,
            'rand_layout_minor': False,
            'rand_layout_full': True,
            'rand_dynamics': True,
        }),
    ]

    # Long episodes because this is a hard environment. You can have up to 10
    # blocks when doing random layouts, and that takes a human 20-30s to
    # process (so 240/8=30s is just enough time to finish a 10-block run if
    # you know what you're doing).
    cluster_ep_len = 240
    cluster_variants = []
    for cluster_cls in (ClusterColourEnv, ClusterShapeEnv):
        cluster_variants.extend([
            (cluster_cls, cluster_ep_len, '-Demo', {
                'rand_shape_colour': False,
                'rand_shape_type': False,
                'rand_layout_minor': False,
                'rand_layout_full': False,
                'rand_shape_count': False,
                'rand_dynamics': False,
            }),
            (cluster_cls, cluster_ep_len, '-TestJitter', {
                'rand_shape_colour': False,
                'rand_shape_type': False,
                'rand_layout_minor': True,
                'rand_layout_full': False,
                'rand_shape_count': False,
                'rand_dynamics': False,
            }),
            (cluster_cls, cluster_ep_len, '-TestColour', {
                'rand_shape_colour': True,
                'rand_shape_type': False,
                'rand_layout_minor': False,
                'rand_layout_full': False,
                'rand_shape_count': False,
                'rand_dynamics': False,
            }),
            (cluster_cls, cluster_ep_len, '-TestShape', {
                'rand_shape_colour': False,
                'rand_shape_type': True,
                'rand_layout_minor': False,
                'rand_layout_full': False,
                'rand_shape_count': False,
                'rand_dynamics': False,
            }),
            (cluster_cls, cluster_ep_len, '-TestLayout', {
                'rand_shape_colour': False,
                'rand_shape_type': False,
                'rand_layout_minor': False,
                'rand_layout_full': True,
                'rand_shape_count': False,
                'rand_dynamics': False,
            }),
            (cluster_cls, cluster_ep_len, '-TestCountPlus', {
                'rand_shape_colour': True,
                'rand_shape_type': True,
                'rand_layout_minor': False,
                'rand_layout_full': True,
                'rand_shape_count': True,
                'rand_dynamics': False,
            }),
            (cluster_cls, cluster_ep_len, '-TestDynamics', {
                'rand_shape_colour': False,
                'rand_shape_type': False,
                'rand_layout_minor': False,
                'rand_layout_full': False,
                'rand_shape_count': False,
                'rand_dynamics': True,
            }),
            (cluster_cls, cluster_ep_len, '-TestAll', {
                'rand_shape_colour': True,
                'rand_shape_type': True,
                'rand_layout_minor': False,
                'rand_layout_full': True,
                'rand_shape_count': True,
                'rand_dynamics': True,
            }),
        ])

    # collection of ALL env specifications
    env_cls_suffix_kwargs = [
        *cluster_variants,
        *find_dupe_variants,
        *fix_colour_variants,
        *make_line_variants,
        *match_regions_variants,
        *move_to_corner_variants,
        *move_to_region_variants,
    ]

    # register all the envs and record their names
    # TODO: make registration lazy, so that I can do it automatically without
    # importing all the benchmark code (including Pyglet code, etc.).
    registered_env_names = []
    for env_class, env_ep_len, env_suffix, env_kwargs in env_cls_suffix_kwargs:
        base_env_name = env_class.make_name(env_suffix)
        registered_env_names.append(base_env_name)
        gym.register(base_env_name,
                     entry_point=env_class,
                     max_episode_steps=env_ep_len,
                     kwargs={
                         'max_episode_steps': env_ep_len,
                         **common_kwargs,
                         **env_kwargs,
                     })

        for preproc_str, constructor in \
                DEFAULT_PREPROC_ENTRY_POINT_WRAPPERS.items():
            new_name = env_class.make_name(env_suffix + f'-{preproc_str}')
            registered_env_names.append(new_name)
            gym.register(new_name,
                         entry_point=constructor(env_class),
                         max_episode_steps=env_ep_len,
                         kwargs={
                             'max_episode_steps': env_ep_len,
                             **common_kwargs,
                             **env_kwargs,
                         })

    train_to_test_map = {}
    observed_demo_envs = set()
    for name in registered_env_names:
        parsed = EnvName(name)
        if parsed.is_test:
            test_l = train_to_test_map.setdefault(parsed.demo_env_name, [])
            test_l.append(parsed.env_name)
        else:
            observed_demo_envs.add(parsed.env_name)

    # use immutable values
    train_to_test_map = {k: tuple(v) for k, v in train_to_test_map.items()}

    envs_with_test_variants = train_to_test_map.keys()
    assert observed_demo_envs == envs_with_test_variants, \
        "there are some train envs without test envs, or test envs without " \
        "train envs"
    sorted_items = sorted(train_to_test_map.items())
    DEMO_ENVS_TO_TEST_ENVS_MAP.update(sorted_items)

    # Debugging environment: MoveToCorner with a nicely shaped reward function
    # that you can simply do RL on.
    debug_mtc_kwargs = dict(max_episode_steps=mtc_ep_len,
                            kwargs={
                                'debug_reward': True,
                                'max_episode_steps': mtc_ep_len,
                                'rand_shape_colour': False,
                                'rand_shape_type': False,
                                'rand_shape_pose': False,
                                'rand_robot_pose': False,
                                **common_kwargs,
                            })
    debug_mtc_suffix = '-DebugReward'
    gym.register(MoveToCornerEnv.make_name(debug_mtc_suffix),
                 entry_point=MoveToCornerEnv,
                 **debug_mtc_kwargs)
    for preproc_str, constructor in \
            DEFAULT_PREPROC_ENTRY_POINT_WRAPPERS.items():
        gym.register(
            MoveToCornerEnv.make_name(f'{debug_mtc_suffix}-{preproc_str}'),
            entry_point=constructor(MoveToCornerEnv),
            **debug_mtc_kwargs)

    return True



import functools

from gym.utils import EzPickle

from magical.base_env import BaseEnv, ez_init
import magical.entities as en
import magical.geom as geom

MIN_REGIONS = 2
MAX_REGIONS = 3
# we have really small goal regions in this domain so we can fit 3+ of them in
# the workspace
MIN_GOAL_SIZE = 0.4
MAX_GOAL_SIZE = 0.5
DEFAULT_ROBOT_POSE = ((0.368, 0.586), 0.718)
DEFAULT_BLOCK_COLOURS = [
    en.ShapeColour.GREEN,
    en.ShapeColour.GREEN,
    en.ShapeColour.BLUE,
]
DEFAULT_BLOCK_SHAPES = [
    en.ShapeType.PENTAGON,
    en.ShapeType.SQUARE,
    en.ShapeType.PENTAGON,
]
DEFAULT_BLOCK_POSES = [
    ((0.289, 0.030), 0.307),
    ((0.133, -0.561), 1.699),
    ((-0.336, 0.000), -1.529),
]
DEFAULT_REGION_XYHWS = [
    (-0.032, 0.348, 0.427, 0.468),
    (0.019, -0.391, 0.460, 0.458),
    (-0.681, 0.196, 0.498, 0.418),
]
DEFAULT_REGION_COLOURS = [
    en.ShapeColour.GREEN,
    en.ShapeColour.GREEN,
    en.ShapeColour.RED,
]


class FixColourEnv(BaseEnv, EzPickle):
    """There are several coloured regions, each containing a single block.
    Exactly one of the coloured regions has a block that _doesn't_ match its
    colour; the block should be removed from this region, but not the
    others."""
    @ez_init()
    def __init__(self,
                 rand_colours=False,
                 rand_shapes=False,
                 rand_count=False,
                 rand_layout_minor=False,
                 rand_layout_full=False,
                 **kwargs):
        super().__init__(**kwargs)
        self.rand_colours = rand_colours
        self.rand_shapes = rand_shapes
        self.rand_count = rand_count
        self.rand_layout_minor = rand_layout_minor
        self.rand_layout_full = rand_layout_full
        if self.rand_count:
            assert self.rand_layout_full and self.rand_shapes \
                and self.rand_colours, "if shape count is randomised then " \
                "layout, shapes, and colours must be fully randomised too"

    @classmethod
    def make_name(cls, suffix=None):
        base_name = 'FixColour'
        return base_name + (suffix or '') + '-v0'

    def on_reset(self):
        robot_pos, robot_angle = DEFAULT_ROBOT_POSE
        robot = self._make_robot(robot_pos, robot_angle)

        block_colours = DEFAULT_BLOCK_COLOURS
        region_colours = DEFAULT_REGION_COLOURS
        block_shapes = DEFAULT_BLOCK_SHAPES
        block_poses = DEFAULT_BLOCK_POSES
        region_xyhws = DEFAULT_REGION_XYHWS

        # randomise count
        n_regions = len(block_colours)
        if self.rand_count:
            n_regions = self.rng.randint(MIN_REGIONS, MAX_REGIONS + 1)
            block_poses = block_poses[:1] * n_regions
            region_xyhws = region_xyhws[:1] * n_regions

        # randomise colours
        if self.rand_colours:
            region_colours = self.rng.choice(en.SHAPE_COLOURS,
                                             size=n_regions).tolist()
            block_colours = list(region_colours)
            # randomly choose one block to be the odd one out
            odd_idx = self.rng.randint(len(block_colours))
            new_col_idx = self.rng.randint(len(en.SHAPE_COLOURS) - 1)
            if en.SHAPE_COLOURS[new_col_idx] == block_colours[odd_idx]:
                new_col_idx += 1
            block_colours[odd_idx] = en.SHAPE_COLOURS[new_col_idx]

        # randomise shapes
        if self.rand_shapes:
            block_shapes = self.rng.choice(en.SHAPE_TYPES,
                                           size=n_regions).tolist()

        # create all regions, randomising their height/width first if necessary
        if self.rand_layout_minor or self.rand_layout_full:
            if self.rand_layout_minor:
                hw_bound = self.JITTER_TARGET_BOUND
            else:
                hw_bound = None
            rand_hw = functools.partial(geom.randomise_hw,
                                        MIN_GOAL_SIZE,
                                        MAX_GOAL_SIZE,
                                        self.rng,
                                        linf_bound=hw_bound)
            region_xyhws = [(x, y, *rand_hw(current_hw=hw))
                            for x, y, *hw in region_xyhws]
        sensors = [
            en.GoalRegion(*xyhw, colour)
            for colour, xyhw in zip(region_colours, region_xyhws)
        ]
        self.add_entities(sensors)
        self._sensors = sensors

        # set up blocks
        blocks = []
        self._target_blocks = []
        for bshape, bcol, tcol, (bpos,
                                 bangle) in zip(block_shapes, block_colours,
                                                region_colours, block_poses):
            new_block = self._make_shape(shape_type=bshape,
                                         colour_name=bcol,
                                         init_pos=bpos,
                                         init_angle=bangle)
            blocks.append(new_block)
            if bcol != tcol:
                # we don't want this region to contain any blocks
                self._target_blocks.append([])
            else:
                # we want this region to keep the original block
                self._target_blocks.append([new_block])
        self.add_entities(blocks)

        # add robot last (for draw order reasons)
        self.add_entities([robot])

        if self.rand_layout_minor or self.rand_layout_full:
            sensors_robot = (*sensors, robot)
            if self.rand_layout_minor:
                pos_limits = self.JITTER_POS_BOUND
                rot_limit = self.JITTER_ROT_BOUND
            else:
                assert self.rand_layout_full
                pos_limits = rot_limit = None
            # don't randomise rotations of goal regions
            rand_rot = [False] * n_regions + [True]
            # don't collision check with blocks
            block_pm_shapes = sum((b.shapes for b in blocks), [])

            # randomise positions of goal regions and robot
            geom.pm_randomise_all_poses(self._space,
                                        sensors_robot,
                                        self.ARENA_BOUNDS_LRBT,
                                        rng=self.rng,
                                        rand_pos=True,
                                        rand_rot=rand_rot,
                                        rel_pos_linf_limits=pos_limits,
                                        rel_rot_limits=rot_limit,
                                        ignore_shapes=block_pm_shapes)

            # shift blocks out of the way of each other
            for block, sensor in zip(blocks, sensors):
                geom.pm_shift_bodies(self._space,
                                     block.bodies,
                                     position=sensor.bodies[0].position)
            # randomise each inner block position separately
            for block, sensor, (_, _,
                                *sensor_hw) in zip(blocks, sensors,
                                                   region_xyhws):
                block_pos_limit = max(0, min(sensor_hw) / 2 - self.SHAPE_RAD)
                if self.rand_layout_minor:
                    block_pos_limit = min(self.JITTER_POS_BOUND,
                                          block_pos_limit)
                geom.pm_randomise_pose(self._space,
                                       block.bodies,
                                       self.ARENA_BOUNDS_LRBT,
                                       ignore_shapes=sensor.shapes,
                                       rng=self.rng,
                                       rand_pos=True,
                                       rand_rot=True,
                                       rel_pos_linf_limit=block_pos_limit,
                                       rel_rot_limit=rot_limit)

        # block lookup index
        self._block_index = en.EntityIndex(blocks)

    def score_on_end_of_traj(self):
        # binary scoring: give score of 1 iff we get exactly the right result
        complete = True
        for sensor, tgt_block_list in zip(self._sensors, self._target_blocks):
            overlap_ents = sensor.get_overlapping_ents(
                com_overlap=True, ent_index=self._block_index)
            if list(overlap_ents) != tgt_block_list:
                complete = False
                break
        return 1.0 if complete else 0.0



"""Tools for working with geometric primitives and randomising aspects of
geometry (shape, pose, etc.)."""

from collections.abc import Iterable, Sequence
import math
import warnings

import numpy as np
import pymunk as pm
from pymunk.vec2d import Vec2d


def regular_poly_circumrad(n_sides, side_length):
    """Circumradius of a regular polygon."""
    return side_length / (2 * math.sin(math.pi / n_sides))


def regular_poly_circ_rad_to_side_length(n_sides, rad):
    """Find side length that gives regular polygon with `n_sides` sides an
    equivalent area to a circle with radius `rad`."""
    p_n = math.pi / n_sides
    return 2 * rad * math.sqrt(p_n * math.tan(p_n))


def regular_poly_apothem_to_side_legnth(n_sides, apothem):
    """Compute side length for regular polygon with given apothem."""
    return 2 * apothem * math.tan(math.pi / n_sides)


def regular_poly_side_length_to_apothem(n_sides, side_length):
    """Compute apothem for regular polygon with given side length."""
    return side_length / (2 * math.tan(math.pi / n_sides))


def compute_regular_poly_verts(n_sides, side_length):
    """Vertices for regular polygon."""
    assert n_sides >= 3
    vertices = []
    step_angle = 2 * math.pi / n_sides
    radius = regular_poly_circumrad(n_sides, side_length)
    first_vertex = pm.vec2d.Vec2d(0, radius)
    for point_num in range(n_sides):
        angle = point_num * step_angle
        vertices.append(first_vertex.rotated(angle))
    vertices = [(v.x, v.y) for v in vertices]
    return vertices


def compute_star_verts(n_points, out_radius, in_radius):
    """Vertices for a star. `n_points` controls the number of points;
    `out_radius` controls distance from points to centre; `in_radius` controls
    radius from "depressions" (the things between points) to centre."""
    assert n_points >= 3
    vertices = []
    out_vertex = pm.vec2d.Vec2d(0, out_radius)
    in_vertex = pm.vec2d.Vec2d(0, in_radius)
    for point_num in range(n_points):
        out_angle = point_num * 2 * math.pi / n_points
        vertices.append(out_vertex.rotated(out_angle))
        in_angle = (2 * point_num + 1) * math.pi / n_points
        vertices.append(in_vertex.rotated(in_angle))
    vertices = [(v.x, v.y) for v in vertices]
    return vertices


def _convert_vec(v):
    if isinstance(v, pm.vec2d.Vec2d):
        return v.x, v.y
    if isinstance(v, (float, int)):
        return (v, v)
    x, y = v
    return (x, y)


def add_vecs(vec1, vec2):
    """Elementwise add vectors represented as vec2ds or tuples or whatever
    (even scalars, in which case they get broadcast). Return result as
    tuple."""
    x1, y1 = _convert_vec(vec1)
    x2, y2 = _convert_vec(vec2)
    return (x1 + x2, y1 + y2)


def mul_vecs(vec1, vec2):
    """Elementwise multiply vectors represented as vec2ds or tuples or
    whatever. Return result as tuple."""
    x1, y1 = _convert_vec(vec1)
    x2, y2 = _convert_vec(vec2)
    return (x1 * x2, y1 * y2)


def rotate_vec(vec, angle):
    # FIXME: this and related functions seem like design mistakes. I should
    # probably be using vec2d.Vec2d everywhere instead of using tuples.
    if not isinstance(vec, pm.vec2d.Vec2d):
        vec = pm.vec2d.Vec2d(*_convert_vec(vec))
    vec_r = vec.rotated(angle)
    return (vec_r.x, vec_r.y)


def rect_verts(w, h):
    # counterclockwise from top right
    return [
        pm.vec2d.Vec2d(w / 2, h / 2),
        pm.vec2d.Vec2d(-w / 2, h / 2),
        pm.vec2d.Vec2d(-w / 2, -h / 2),
        pm.vec2d.Vec2d(w / 2, -h / 2),
    ]


class PlacementError(Exception):
    """Raised when `pm_randomise_pose` cannot find an appropriate
    (non-colliding) pose for the given object."""
    pass


def pm_randomise_pose(space,
                      bodies,
                      arena_lrbt,
                      rng,
                      rand_pos=True,
                      rand_rot=True,
                      rel_pos_linf_limit=None,
                      rel_rot_limit=None,
                      ignore_shapes=None,
                      rejection_tests=()):
    r"""Do rejection sampling to choose a position and/or orientation which
    ensures the given bodies and their attached shapes do not collide with any
    other collidable shape in the space, while still falling entirely within
    arena_xyhw. Note that position/orientation will be chosen in terms of the
    first body in the given list of bodies, then later bodies attached to it
    will be repositioned and rotated accordingly.

    Args:
        space (pm.Space): the space to place the given bodies in.
        bodies ([pm.Body]): a list of bodies to place. They should maintain the
            same relative positions and orientations. Usually you'll only need
            to pass one body, although passing multiple bodies can be useful
            when the bodies have a pin joint (e.g. the robot's body is attached
            to its fingers this way).
        arena_lrbt ([int]): bounding box to place the bodies in.
        rand_pos (bool or [bool]): should position be randomised? (optionally
            specified separately for each entity)
        rand_rot (bool or [bool]): should rotation be randomised? (optionally
            specified for each entity)
        rel_pos_linf_limit (float or [float]): bound on the $\ell_\infty$
            distance between new sampled position and original position.
            (optionally specified per-entity)
        rel_rot_limit (float or [float]): maximum difference (in radians)
            between original main body orientation and new main body
            orientation. (optionally per-entity)
        rejection_tests ([(locals()) -> bool]): additional rejection tests to
            apply. If any one of these functions returns "True", then the shape
            pose will be rejected and re-sampled. Useful for, e.g., ensuring
            that placed shapes do not coincide with certain existing objects.

    Returns (int): number of random placements attempted before finding a
        successful one."""
    assert rand_pos or rand_rot, \
        "need to randomise at least one thing, or placement may be impossible"
    assert len(bodies) >= 1, "no bodies given (?)"
    main_body = bodies[0]

    # Need to compute coordinates of other bodies in frame of main body
    saved_positions = [Vec2d(body.position) for body in bodies]
    saved_angles = [float(body.angle) for body in bodies]
    orig_main_angle = float(main_body.angle)
    orig_main_pos = Vec2d(main_body.position)
    # local_pos_offsets = [
    #     (body.position - orig_main_pos).rotated(-orig_main_angle)
    #     for body in bodies
    # ]
    # local_angle_deltas = [body.angle - orig_main_angle for body in bodies]

    shape_set = set()
    for body in bodies:
        shape_set.update(body.shapes)

    if ignore_shapes is not None:
        ignore_set = set(ignore_shapes)
    else:
        ignore_set = set()

    arena_l, arena_r, arena_b, arena_t = arena_lrbt

    if rel_pos_linf_limit is not None:
        assert 0 <= rel_pos_linf_limit
        init_x, init_y = main_body.position
        pos_x_minmax = (max(arena_l, init_x - rel_pos_linf_limit),
                        min(arena_r, init_x + rel_pos_linf_limit))
        pos_y_minmax = (max(arena_b, init_y - rel_pos_linf_limit),
                        min(arena_t, init_y + rel_pos_linf_limit))
    else:
        pos_x_minmax = (arena_l, arena_r)
        pos_y_minmax = (arena_b, arena_t)

    if rel_rot_limit is not None:
        assert 0 <= rel_rot_limit
        rot_min = orig_main_angle - rel_rot_limit
        rot_max = orig_main_angle + rel_rot_limit
    else:
        rot_min = -np.pi
        rot_max = np.pi

    # If we exceed this many tries then fitting is probably impossible, or
    # impractically hard. We'll warn if we get anywhere close to that number.
    max_tries = 10000
    warn_tries = int(max_tries / 10)

    n_tries = 0
    while n_tries < max_tries:
        # generate random position
        if rand_pos:
            new_main_body_pos = Vec2d(rng.uniform(*pos_x_minmax),
                                      rng.uniform(*pos_y_minmax))
        else:
            new_main_body_pos = orig_main_pos

        # generate random orientation
        if rand_rot:
            new_angle = rng.uniform(rot_min, rot_max)
        else:
            new_angle = orig_main_angle

        # apply new position/orientation to all bodies
        pm_shift_bodies(space,
                        bodies,
                        position=new_main_body_pos,
                        angle=new_angle)

        # apply collision tests
        reject = False
        for shape in shape_set:
            query_result = space.shape_query(shape)
            collisions = set(r.shape for r in query_result) - ignore_set
            # reject if we have any (non-self-)collisions
            if len(collisions) > 0:
                reject = True
                break

        # apply custom rejection tests, if any
        if not reject:
            for rejection_test in rejection_tests:
                reject = reject or rejection_test(locals())
                if reject:
                    break

        if not reject:
            # if we get to here without rejecting, then this is a good
            # orientation
            break

        n_tries += 1
    else:
        # reset original positions before raising exception
        for body, saved_pos, saved_angle in zip(bodies, saved_positions,
                                                saved_angles):
            body.position = saved_pos
            body.angle = saved_angle
            space.reindex_shapes_for_body(body)

        raise PlacementError(
            f"Could not place bodies {bodies} in space {space} after "
            f"{n_tries} attempts. rand_pos={rand_pos}, rand_rot={rand_rot}, "
            f"arena_lrbt={arena_lrbt}.")

    if n_tries > warn_tries:
        warnings.warn(f"Took {n_tries}>{warn_tries} samples to place shape.")

    return n_tries


def _listify(value, n):
    if isinstance(value, Iterable):
        # if `value` is already an iterable, then cast it to a sequence type
        # and return it
        if not isinstance(value, Sequence):
            rv = list(value)
        else:
            rv = value
        assert len(rv) == n, (len(rv), n)
        return rv
    # otherwise, duplicate value `n` times
    return [value] * n


def pm_randomise_all_poses(space,
                           entities,
                           arena_lrbt,
                           rng,
                           rand_pos=True,
                           rand_rot=True,
                           rel_pos_linf_limits=None,
                           rel_rot_limits=None,
                           ignore_shapes=None,
                           max_retries=10,
                           rejection_tests=()):
    """Randomise poses of *all* entities in the given list of entities."""
    # create placeholder limits if necessary
    nents = len(entities)
    rel_pos_linf_limits = _listify(rel_pos_linf_limits, nents)
    rel_rot_limits = _listify(rel_rot_limits, nents)
    rand_pos = _listify(rand_pos, nents)
    rand_rot = _listify(rand_rot, nents)

    for retry in range(max_retries):
        # disable collisions for all entities
        ent_filters = []
        for entity in entities:
            shape_filters = []
            for s in entity.shapes:
                shape_filters.append(s.filter)
                # categories=0 makes it collide with nothing
                s.filter = s.filter._replace(categories=0)
            ent_filters.append(shape_filters)

        for (entity, shape_filters, pos_limit, rot_limit, should_rand_pos,
             should_rand_rot) in zip(entities, ent_filters,
                                     rel_pos_linf_limits, rel_rot_limits,
                                     rand_pos, rand_rot):
            # re-enable collisions for this entity (previous entities will
            # already have collisions enabled, and later entities will still
            # have collisions disabled)
            for s, filt in zip(entity.shapes, shape_filters):
                s.filter = filt

            # now randomise pose, avoiding entities that have previously been
            # placed or which are not in the supplied list
            try:
                pm_randomise_pose(space,
                                  entity.bodies,
                                  arena_lrbt,
                                  rng,
                                  rand_pos=should_rand_pos,
                                  rand_rot=should_rand_rot,
                                  rel_pos_linf_limit=pos_limit,
                                  rel_rot_limit=rot_limit,
                                  ignore_shapes=ignore_shapes,
                                  rejection_tests=rejection_tests)
            except PlacementError as ex:
                if retry == max_retries - 1:
                    raise
                print(f"Got PlacementError ({ex}) on retry {retry + 1}"
                      f"/{max_retries}, restarting")
                break
        else:
            break


def randomise_hw(min_side, max_side, rng, current_hw=None, linf_bound=None):
    """Randomise height and width parameters within some supplied bounds.
    Useful for randomising goal region height/width in a reasonably uniform
    way."""
    assert min_side <= max_side
    minima = np.asarray((min_side, min_side))
    maxima = np.asarray((max_side, max_side))
    if linf_bound is not None:
        assert linf_bound == float(linf_bound)
        assert current_hw is not None
        assert len(current_hw) == 2
        current_hw = np.asarray(current_hw)
        minima = np.maximum(minima, current_hw - linf_bound)
        maxima = np.minimum(maxima, current_hw + linf_bound)
    h, w = rng.uniform(minima, maxima)
    return h, w


def pm_shift_bodies(space, bodies, position=None, angle=None):
    """Apply a rigid transform to the given bodies to move them into the given
    position and/or angle. Note that position and angle are specified for the
    first body only; later bodies are modelled as if attached to the first."""
    assert len(bodies) >= 1

    root_angle = bodies[0].angle
    root_position = bodies[0].position
    if angle is None:
        angle = root_angle
    if position is None:
        position = root_position

    # cast to right types
    position = pm.Vec2d(position)
    angle = float(angle)

    for body in bodies:
        local_angle_delta = body.angle - root_angle
        local_pos_delta = body.position - root_position
        body.angle = angle + local_angle_delta
        body.position = position + local_pos_delta.rotated(angle - root_angle)
        space.reindex_shapes_for_body(body)



# This file was taken from Pyglet 2.0dev0 at
# 94bf013cc76312ed91cbeedd81f4bde1b9bb9273

# ----------------------------------------------------------------------------
# pyglet
# Copyright (c) 2006-2008 Alex Holkner
# Copyright (c) 2008-2019 pyglet contributors
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#  * Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#  * Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in
#    the documentation and/or other materials provided with the
#    distribution.
#  * Neither the name of pyglet nor the names of its
#    contributors may be used to endorse or promote products
#    derived from this software without specific prior written
#    permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# ----------------------------------------------------------------------------
from pyglet.gl import *


def get_max_color_attachments():
    """Get the maximum allow Framebuffer Color attachements"""
    number = GLint()
    glGetIntegerv(GL_MAX_COLOR_ATTACHMENTS, number)
    return number.value


class Renderbuffer:
    """OpenGL Renderbuffer Object"""

    def __init__(self, width, height, internal_format, samples=1):
        """Create an instance of a Renderbuffer object."""
        self._id = GLuint()
        self._width = width
        self._height = height
        self._internal_format = internal_format

        glGenRenderbuffers(1, self._id)
        glBindRenderbuffer(GL_RENDERBUFFER, self._id)

        if samples > 1:
            glRenderbufferStorageMultisample(GL_RENDERBUFFER, samples, internal_format, width, height)
        else:
            glRenderbufferStorage(GL_RENDERBUFFER, internal_format, width, height)

        glBindRenderbuffer(GL_RENDERBUFFER, 0)

    @property
    def id(self):
        return self._id.value

    @property
    def width(self):
        return self._width

    @property
    def height(self):
        return self._height

    def bind(self):
        glBindRenderbuffer(GL_RENDERBUFFER, self._id)

    @staticmethod
    def unbind():
        glBindRenderbuffer(GL_RENDERBUFFER, 0)

    def delete(self):
        glDeleteRenderbuffers(1, self._id)

    def __del__(self):
        try:
            glDeleteRenderbuffers(1, self._id)
        # Python interpreter is shutting down:
        except ImportError:
            pass

    def __repr__(self):
        return "{}(id={})".format(self.__class__.__name__, self._id.value)


class Framebuffer:
    """OpenGL Framebuffer Object"""

    _max_color_attachments = get_max_color_attachments()

    def __init__(self, target=GL_FRAMEBUFFER):
        """Create an OpenGL Framebuffer object.

        :rtype: :py:class:`~pyglet.image.Framebuffer`

        .. versionadded:: 2.0
        """
        self._id = GLuint()
        glGenFramebuffers(1, self._id)
        self._attachment_types = 0
        self._width = 0
        self._height = 0

    @property
    def id(self):
        return self._id.value

    @property
    def width(self):
        """The width of the widest attachment."""
        return self._width

    @property
    def height(self):
        """The width of the widest attachment."""
        return self._height

    def bind(self, target=GL_FRAMEBUFFER):
        glBindFramebuffer(target, self._id)

    @staticmethod
    def unbind(target=GL_FRAMEBUFFER):
        glBindFramebuffer(target, 0)

    def clear(self):
        if self._attachment_types:
            self.bind()
            glClear(self._attachment_types)
            self.unbind()

    def delete(self):
        glDeleteFramebuffers(1, self._id)

    @property
    def is_complete(self):
        return glCheckFramebufferStatus(GL_FRAMEBUFFER) == GL_FRAMEBUFFER_COMPLETE

    @staticmethod
    def get_status():
        states = {GL_FRAMEBUFFER_UNSUPPORTED: "Framebuffer unsupported. Try another format.",
                  GL_FRAMEBUFFER_INCOMPLETE_ATTACHMENT: "Framebuffer incomplete attachment.",
                  GL_FRAMEBUFFER_INCOMPLETE_MISSING_ATTACHMENT: "Framebuffer missing attachment.",
                  GL_FRAMEBUFFER_INCOMPLETE_DIMENSIONS_EXT: "Framebuffer unsupported dimension.",
                  GL_FRAMEBUFFER_INCOMPLETE_FORMATS_EXT: "Framebuffer incomplete formats.",
                  GL_FRAMEBUFFER_INCOMPLETE_DRAW_BUFFER: "Framebuffer incomplete draw buffer.",
                  GL_FRAMEBUFFER_INCOMPLETE_READ_BUFFER: "Framebuffer incomplete read buffer.",
                  GL_FRAMEBUFFER_COMPLETE: "Framebuffer is complete."}

        gl_status = glCheckFramebufferStatus(GL_FRAMEBUFFER)

        return states.get(gl_status, "Unknown error")

    def attach_texture(self, target, attachment, texture):
        """Attach a Texture to the Framebuffer

        :Parameters:
            `target` : int
                Specifies the framebuffer target. target must be GL_DRAW_FRAMEBUFFER,
                GL_READ_FRAMEBUFFER, or GL_FRAMEBUFFER. GL_FRAMEBUFFER is equivalent
                to GL_DRAW_FRAMEBUFFER.
            `attachment` : int
                Specifies the attachment point of the framebuffer. attachment must be
                GL_COLOR_ATTACHMENTi, GL_DEPTH_ATTACHMENT, GL_STENCIL_ATTACHMENT or
                GL_DEPTH_STENCIL_ATTACHMENT.
            `texture` : pyglet.image.Texture
                Specifies the texture object to attach to the framebuffer attachment
                point named by attachment.
        """
        self.bind()
        glFramebufferTexture2D(target, attachment, texture.target, texture.id, texture.level)
        # glFramebufferTexture2D(target, attachment, texture.target, texture.id, texture.level)
        self._attachment_types |= attachment
        self._width = max(texture.width, self._width)
        self._height = max(texture.height, self._height)
        self.unbind()

    def attach_texture_layer(self, target, attachment, texture, level, layer):
        """Attach a Texture layer to the Framebuffer

        :Parameters:
            `target` : int
                Specifies the framebuffer target. target must be GL_DRAW_FRAMEBUFFER,
                GL_READ_FRAMEBUFFER, or GL_FRAMEBUFFER. GL_FRAMEBUFFER is equivalent
                to GL_DRAW_FRAMEBUFFER.
            `attachment` : int
                Specifies the attachment point of the framebuffer. attachment must be
                GL_COLOR_ATTACHMENTi, GL_DEPTH_ATTACHMENT, GL_STENCIL_ATTACHMENT or
                GL_DEPTH_STENCIL_ATTACHMENT.
            `texture` : pyglet.image.TextureArray
                Specifies the texture object to attach to the framebuffer attachment
                point named by attachment.
            `level` : int
                Specifies the mipmap level of texture to attach.
            `layer` : int
                Specifies the layer of texture to attach.
        """
        self.bind()
        glFramebufferTextureLayer(target, attachment, texture.id, level, layer)
        self._attachment_types |= attachment
        self._width = max(texture.width, self._width)
        self._height = max(texture.height, self._height)
        self.unbind()

    def attach_renderbuffer(self, target, attachment, renderbuffer):
        """"Attach a Renderbuffer to the Framebuffer

        :Parameters:
            `target` : int
                Specifies the framebuffer target. target must be GL_DRAW_FRAMEBUFFER,
                GL_READ_FRAMEBUFFER, or GL_FRAMEBUFFER. GL_FRAMEBUFFER is equivalent
                to GL_DRAW_FRAMEBUFFER.
            `attachment` : int
                Specifies the attachment point of the framebuffer. attachment must be
                GL_COLOR_ATTACHMENTi, GL_DEPTH_ATTACHMENT, GL_STENCIL_ATTACHMENT or
                GL_DEPTH_STENCIL_ATTACHMENT.
            `renderbuffer` : pyglet.image.Renderbuffer
                Specifies the Renderbuffer to attach to the framebuffer attachment
                point named by attachment.
        """
        self.bind()
        glFramebufferRenderbuffer(target, attachment, GL_RENDERBUFFER, renderbuffer.id)
        self._attachment_types |= attachment
        self._width = max(renderbuffer.width, self._width)
        self._height = max(renderbuffer.height, self._height)
        self.unbind()

    def __del__(self):
        try:
            glDeleteFramebuffers(1, self._id)
        # Python interpreter is shutting down:
        except ImportError:
            pass

    def __repr__(self):
        return "{}(id={})".format(self.__class__.__name__, self._id.value)



"""Tool for demonstrating MAGICAL and collecting demos."""

import collections
import datetime
import gzip
import os
import sys
import time

import click
import cloudpickle
import gym
import numpy as np
from pyglet.window import key

from magical.benchmarks import register_envs
from magical.entities import RobotAction as RA
from magical.saved_trajectories import MAGICALTrajectory


def get_unique_fn(env_name):
    now = datetime.datetime.now()
    time_str = now.strftime('%FT%H:%M:%S')
    return f"demo-{env_name}-{time_str}.pkl.gz"


class Accumulator:
    """Accumulates trajectory steps."""
    def __init__(self):
        self.partial_traj = []

    def add_step(self, step_dict):
        self.partial_traj.append(step_dict)

    def finish_trajectory(self):
        out_unstacked = collections.defaultdict(list)
        for step_d in self.partial_traj:
            for k, v in step_d.items():
                out_unstacked[k].append(v)
        out_stacked = {
            k: np.stack(v, axis=0)
            for k, v in out_unstacked.items()
        }
        self.partial_traj = []
        return MAGICALTrajectory(**out_stacked)


@click.command()
@click.option("--record",
              type=str,
              default=None,
              help="directory to record demos to, if any")
@click.option("--env-name",
              default='MoveToCorner-Demo-LoResStack-v0',
              help='name of environment')
@click.option("--print-spec/--no-print-spec",
              default=False,
              help="print env spec?")
def main(record, env_name, print_spec):
    if record:
        record_dir = os.path.abspath(record)
        print(f"Will record demos to '{record_dir}'")
        os.makedirs(record_dir, exist_ok=True)
        traj_accum = Accumulator()

    register_envs()
    env = gym.make(env_name)
    try:
        obs = env.reset()
        if print_spec:
            print('\n')
            env.debug_print_entity_spec()
        was_done_on_prev_step = False
        if record:
            traj_accum.add_step({"obs": obs})
            started = False
            print("Will only start recording on first key press")
        else:
            started = True

        # first render to open window
        env.render(mode='human')

        # keys that are depressed will end up in key_map
        key_map = key.KeyStateHandler()
        env.viewer.window.push_handlers(key_map)

        # render loop
        spf = 1.0 / env.fps
        while env.viewer.isopen:
            if key_map[key.R]:
                # drop traj and don't save
                obs = env.reset()
                if print_spec:
                    print('\n')
                    env.debug_print_entity_spec()
                if record:
                    started = False
                    traj_accum = Accumulator()
                    traj_accum.add_step({"obs": obs})
                else:
                    started = True
                was_done_on_prev_step = False

            # for limiting FPS
            frame_start = time.time()

            # movement and gripper keys
            act_flags = [RA.NONE, RA.NONE, RA.OPEN]
            if key_map[key.UP] and not key_map[key.DOWN]:
                act_flags[0] = RA.UP
            elif key_map[key.DOWN] and not key_map[key.UP]:
                act_flags[0] = RA.DOWN
            if key_map[key.LEFT] and not key_map[key.RIGHT]:
                act_flags[1] = RA.LEFT
            elif key_map[key.RIGHT] and not key_map[key.LEFT]:
                act_flags[1] = RA.RIGHT
            if key_map[key.SPACE]:
                # close gripper (otherwise it's open)
                act_flags[2] = RA.CLOSE
            # "flat" integer action
            action = env.flags_to_action(act_flags)
            was_started = started
            started = started or action != 0
            if started and not was_started:
                print("Detected non-null action, starting recording")

            if started:
                obs, rew, done, info = env.step(action)

                if done and not was_done_on_prev_step:
                    print(f"Done, score {info['eval_score']:.4g}/1.0")

                if record:
                    traj_accum.add_step({
                        "rews": rew,
                        "obs": obs,
                        "acts": action,
                        "infos": info,
                    })
                    if done and not was_done_on_prev_step:
                        traj = traj_accum.finish_trajectory()
                        new_path = os.path.join(record_dir,
                                                get_unique_fn(env_name))
                        pickle_data = {
                            'env_name': env_name,
                            'trajectory': traj,
                            'score': info['eval_score'],
                        }
                        print(f"Saving trajectory ({len(traj.obs)} obs, "
                              f"{len(traj.acts)} actions, {len(traj.rews)} "
                              f"rews) to '{new_path}'")
                        with gzip.GzipFile(new_path, 'wb') as fp:
                            cloudpickle.dump(pickle_data, fp)

                # for things we only want to run the FIRST time the env gives
                # is a 'done' flag
                was_done_on_prev_step = done

            # render to screen
            env.render(mode='human')

            # wait for next frame
            elapsed = time.time() - frame_start
            if elapsed < spf:
                time.sleep(spf - elapsed)
    finally:
        # once done, close the window
        env.close()


if __name__ == '__main__':
    try:
        with main.make_context(sys.argv[0], sys.argv[1:]) as ctx:
            result = main.invoke(ctx)
    except click.ClickException as e:
        e.show()
        sys.exit(e.exit_code)
    except click.exceptions.Exit as e:
        sys.exit(e.exit_code)



"""Common visual settings and colour tools for this benchmark."""

import colorsys


def rgb(r, g, b):
    return (r / 255.0, g / 255.0, b / 255.0)


def darken_rgb(rgb):
    """Produce a darker version of a base colour."""
    h, l, s = colorsys.rgb_to_hls(*rgb)
    hls_new = (h, max(0, l * 0.9), s)
    return colorsys.hls_to_rgb(*hls_new)


def lighten_rgb(rgb, times=1):
    """Produce a lighter version of a given base colour."""
    h, l, s = colorsys.rgb_to_hls(*rgb)
    mult = 1.4 ** times
    hls_new = (h, 1 - (1 - l) / mult, s)
    return colorsys.hls_to_rgb(*hls_new)


GOAL_LINE_THICKNESS = 0.01
SHAPE_LINE_THICKNESS = 0.015
ROBOT_LINE_THICKNESS = 0.01
COLOURS_RGB = {
    # Original colours from Wikipedia. They were all pastel base colours useful
    # for shape bodies; can use darken_rgb to get highlights (e.g. borders) and
    # lighten_rgb to get get fades useful for backgrounds.

    # Unfortunately pastel RGBY originally looked too much like Google or
    # Microsoft's corporate brands. I tried to replace them with
    # Berkeley-branded versions of these colours from
    # https://brand.berkeley.edu/colors/ (possibly lightened).

    # 'red': rgb(255, 192, 203),  # pink
    # 'green': rgb(170, 240, 209),  # magic mint
    # 'blue': rgb(137, 207, 240),  # baby blue
    # 'yellow': rgb(255, 229, 180),  # peach

    'blue': lighten_rgb(rgb(0x3B, 0x7E, 0xA1), 1.7),  # founder's rock
    'yellow': lighten_rgb(rgb(0xFD, 0xB5, 0x15), 1.7),  # california gold
    'red': lighten_rgb(rgb(0xEE, 0x1F, 0x60), 1.7),  # rose garden
    'green': lighten_rgb(rgb(0x85, 0x94, 0x38), 1.7),  # soybean

    'grey': rgb(162, 163, 175),  # cool grey (not sure which one)
    'brown': rgb(224, 171, 118),  # buff
}
# "zoom out" factor when rendering arena; values above 1 will show parts of the
# arena border
ARENA_ZOOM_OUT = 1.02



"""Gym wrapper for shape-pushing environment."""

import abc
import collections
import functools
import inspect

import gym
from gym import spaces
from gym.utils import EzPickle
import numpy as np
import pymunk as pm

import magical.entities as en
import magical.gym_render as r
from magical.phys_vars import PhysicsVariablesBase, PhysVar
from magical.style import ARENA_ZOOM_OUT, COLOURS_RGB, lighten_rgb


def ez_init(*args, **kwargs):
    """Decorator to initialise EzPickle from all arguments and keyword
    arguments. Use it like this:

        class C(…, EzPickle):
            @ez_init()
            def __init__(self, spam, ham, …):
                …"""
    assert len(args) == 0 and len(kwargs) == 0, \
        "ez_init takes no args at the moment (use `@ez_init()` only)"

    def inner_decorator(method):
        # sanity checks
        msg = f"Got function {method}; should only be used to decorate " \
              f"__init__() methods of classes"
        assert inspect.isfunction(method), msg
        assert method.__name__ == '__init__', msg

        @functools.wraps(method)
        def wrapper(*args, **kwargs):
            self_var = args[0]
            EzPickle.__init__(self_var, *args[1:], **kwargs)
            return method(*args, **kwargs)

        return wrapper

    return inner_decorator


class PhysicsVariables(PhysicsVariablesBase):
    """Default values & randomisation ranges for key physical parameters of the
    environment."""
    # robot_pos_joint_max_force = PhysVar(3, (2.5, 4))
    robot_pos_joint_max_force = PhysVar(3, (2.2, 3.5))
    robot_rot_joint_max_force = PhysVar(1, (0.7, 1.5))
    robot_finger_max_force = PhysVar(4, (2.5, 4.5))
    shape_trans_joint_max_force = PhysVar(1.5, (1.0, 1.8))
    shape_rot_joint_max_force = PhysVar(0.1, (0.07, 0.15))


class BaseEnv(gym.Env, abc.ABC):
    # constants for all envs
    ROBOT_RAD = 0.2
    ROBOT_MASS = 1.0
    SHAPE_RAD = ROBOT_RAD * 0.6
    ARENA_BOUNDS_LRBT = [-1, 1, -1, 1]
    ARENA_SIZE_MAX = max(ARENA_BOUNDS_LRBT)
    # minimum and maximum size of goal regions used during randomisation
    RAND_GOAL_MIN_SIZE = 0.5
    RAND_GOAL_MAX_SIZE = 0.8
    RAND_GOAL_SIZE_RANGE = RAND_GOAL_MAX_SIZE - RAND_GOAL_MIN_SIZE
    # the following are used to standardise what "jitter" means across
    # different tasks
    JITTER_PCT = 0.05
    JITTER_POS_BOUND = ARENA_SIZE_MAX * JITTER_PCT / 2.0
    JITTER_ROT_BOUND = JITTER_PCT * np.pi
    JITTER_TARGET_BOUND = JITTER_PCT * RAND_GOAL_SIZE_RANGE / 2

    def __init__(self,
                 *,
                 res_hw=(256, 256),
                 fps=20,
                 phys_steps=10,
                 phys_iter=10,
                 max_episode_steps=None,
                 rand_dynamics=False,
                 ego_view=True,
                 allo_view=True):
        self.phys_iter = phys_iter
        self.phys_steps = phys_steps
        self.fps = fps
        self.res_hw = res_hw
        self.max_episode_steps = max_episode_steps
        self.ego_view = ego_view
        self.allo_view = allo_view
        assert self.ego_view or self.allo_view, \
            "must use egocentric view or allocentric view (or both)"
        make_image_space = functools.partial(spaces.Box,
                                             low=0.0,
                                             high=255,
                                             shape=(*res_hw, 3),
                                             dtype=np.uint8)
        space_dict = collections.OrderedDict()
        if self.allo_view:
            space_dict['allo'] = make_image_space()
        if self.ego_view:
            space_dict['ego'] = make_image_space()
        self.observation_space = spaces.Dict(space_dict)
        # action space includes every combination of those
        self.action_space = spaces.Discrete(len(en.ACTION_NUMS_FLAGS_NAMES))

        # state/rendering (see reset())
        self._entities = None
        self._space = None
        self._robot = None
        self._episode_steps = None
        self._phys_vars = None
        # this is for background rendering
        self.viewer = None
        # common randomisation option for all envs
        self.rand_dynamics = rand_dynamics

        self.seed()

    def action_to_flags(self, int_action):
        """Parse a 'flat' integer action into a combination of flags."""
        return en.ACTION_ID_TO_FLAGS[int(int_action)]

    def flags_to_action(self, flags):
        """Convert a 'structured' list of flags into a single flag integer
        action. Inverse of action_to_flags."""
        return en.FLAGS_TO_ACTION_ID[tuple(flags)]

    def seed(self, seed=None):
        """Initialise the PRNG and return seed necessary to reproduce results.

        (TODO: should I also seed action/observation spaces? Not clear.)"""
        if seed is None:
            seed = np.random.randint(0, (1 << 31) - 1)
        self.rng = np.random.RandomState(seed=seed)
        return [seed]

    def _make_robot(self, init_pos, init_angle):
        return en.Robot(radius=self.ROBOT_RAD,
                        init_pos=init_pos,
                        init_angle=init_angle,
                        mass=self.ROBOT_MASS)

    def _make_shape(self, **kwargs):
        return en.Shape(shape_size=self.SHAPE_RAD, **kwargs)

    @classmethod
    @abc.abstractmethod
    def make_name(cls, suffix=None):
        """Return a name for an env based on this one, but using the supplied
        suffix. For instance, if an environment were called 'CircleMove' and
        its version were v0, then env_cls.make_name('Hard') would return
        'CircleMoveHard-v0'. If no suffix is supplied then it will just return
        the base name with a version.

        Args:
            suffix (str): the suffix to append to the base name for this
            env.

        Returns:
            name (str): full, Gym-compatible name for this env, with the
                included name suffix."""
    @abc.abstractmethod
    def on_reset(self):
        """Set up entities necessary for this environment, and reset any other
        data needed for the env. Must create a robot in addition to any
        necessary entities, and return bot hthe robot and the other entities
        separately.

        Returns: a tuple with the following elements:
            robot (en.Robot): an initialised robot to be controlled by the
                user.
            ents ([en.Entity]): list of other entities necessary for this
                environment."""
    def add_entities(self, entities):
        """Adds a list of entities to the current entities list and sets it up.
        Only intended to be used from within on_reset(). Needs to be called for
        every created entity or else they will not be added to the space!

        Args:
            entities (en.Entity): the entity to add."""
        for entity in entities:
            if isinstance(entity, en.Robot):
                self._robot = entity
            self._entities.append(entity)
            entity.setup(self.viewer, self._space, self._phys_vars)

    def reset(self):
        self._episode_steps = 0
        # delete old entities/space
        self._entities = []
        self._space = None
        self._robot = None
        self._phys_vars = None
        if self.viewer is None:
            res_h, res_w = self.res_hw
            background_colour = lighten_rgb(COLOURS_RGB['grey'], times=4)
            self.viewer = r.Viewer(res_w,
                                   res_h,
                                   visible=False,
                                   background_rgb=background_colour)
        else:
            # these will get added back later
            self.viewer.reset_geoms()
        self._space = pm.Space()
        self._space.collision_slop = 0.01
        self._space.iterations = self.phys_iter

        if self.rand_dynamics:
            # Randomise the physics properties of objects and the robot a
            # little bit.
            self._phys_vars = PhysicsVariables.sample(self.rng)
        else:
            self._phys_vars = PhysicsVariables.defaults()

        # set up robot and arena
        arena_l, arena_r, arena_b, arena_t = self.ARENA_BOUNDS_LRBT
        self._arena = en.ArenaBoundaries(left=arena_l,
                                         right=arena_r,
                                         bottom=arena_b,
                                         top=arena_t)
        self._arena_w = arena_r - arena_l
        self._arena_h = arena_t - arena_b
        self.add_entities([self._arena])
        reset_rv = self.on_reset()
        assert reset_rv is None, \
            f"on_reset method of {type(self)} returned {reset_rv}, but "\
            f"should return None"
        assert isinstance(self._robot, en.Robot)
        assert len(self._entities) >= 1

        assert np.allclose(self._arena.left + self._arena.right, 0)
        assert np.allclose(self._arena.bottom + self._arena.top, 0)
        self._use_allo_cam()

        # # step forward by one second so PyMunk can recover from bad initial
        # # conditions
        # # (disabled some time in 2019; don't think it's necessary with
        # # well-designed envs)
        # forward_time = 1.0
        # forward_frames = int(math.ceil(forward_time * self.fps))
        # for _ in range(forward_frames):
        #     self._phys_steps_on_frame()

        return self.render(mode='rgb_array')

    def _phys_steps_on_frame(self):
        phys_steps = 10
        spf = 1 / self.fps
        dt = spf / phys_steps
        for i in range(phys_steps):
            for ent in self._entities:
                ent.update(dt)
            self._space.step(dt)

    @abc.abstractmethod
    def score_on_end_of_traj(self):
        """Compute the score for this trajectory. Only called at the last step
        of the trajectory.

        Returns:
           score (float): number in [0, 1] indicating the worst possible
               performance (0), the best possible performance (1) or something
               in between. Should apply to the WHOLE trajectory."""
    def step(self, action):
        # step forward physics
        ac_flag_ud, ac_flag_lr, ac_flag_grip = self.action_to_flags(action)
        action_flag = en.RobotAction.NONE
        action_flag |= ac_flag_ud
        action_flag |= ac_flag_lr
        action_flag |= ac_flag_grip
        self._robot.set_action(action_flag)
        self._phys_steps_on_frame()

        info = {}
        # always 0 reward (avoids training RL algs accidentally)
        reward = 0.0

        # check episode step count
        self._episode_steps += 1
        done = False
        eval_score = 0.0
        if self.max_episode_steps is not None:
            done = done or self._episode_steps >= self.max_episode_steps
        if done:
            eval_score = self.score_on_end_of_traj()
            assert 0 <= eval_score <= 1, \
                f'eval score {eval_score} out of range for env {self}'
            # These were my attempts at sneaking episode termination info
            # through the Monitor and SubprocVecEnv wrappers ('monitor_info'
            # was a keyword I gave to info_keywords in the Monitor
            # constructor). I don't know why, but neither approach worked.
            # info['monitor_info'] = end_ep_dict
            # info['episode'] = end_ep_dict
        # we *always* include a score, even if it's zero, because at least one
        # RL framework (rlpyt) refuses to recognise keys that aren't present at
        # the first time step.
        info.update(eval_score=eval_score)

        obs_dict_u8 = self.render(mode='rgb_array')

        return obs_dict_u8, reward, done, info

    def _use_ego_cam(self):
        self.viewer.set_cam_follow(
            source_xy_world=(self._robot.robot_body.position.x,
                             self._robot.robot_body.position.y),
            target_xy_01=(0.5, 0.15),
            viewport_hw_world=(self._arena_h * ARENA_ZOOM_OUT,
                               self._arena_w * ARENA_ZOOM_OUT),
            rotation=self._robot.robot_body.angle)

    def _use_allo_cam(self):
        self.viewer.set_bounds(left=self._arena.left * ARENA_ZOOM_OUT,
                               right=self._arena.right * ARENA_ZOOM_OUT,
                               bottom=self._arena.bottom * ARENA_ZOOM_OUT,
                               top=self._arena.top * ARENA_ZOOM_OUT)

    def render(self, mode='human'):
        # FIXME: would be simpler for this to special-case to mode='human',
        # mode='ego', and mode='allo', then handle the logic for combining ego
        # + allo in .step().
        if self.viewer is None:
            return None

        for ent in self._entities:
            ent.pre_draw()

        if mode == 'human':
            self.viewer.window.set_visible(True)
        else:
            assert mode == 'rgb_array'

        view_dict = collections.OrderedDict()
        is_human = mode == 'human'
        if self.allo_view:
            self._use_allo_cam()
            view_dict['allo'] = self.viewer.render(return_rgb_array=True,
                                                   update_foreground=is_human)
        if self.ego_view:
            self._use_ego_cam()
            # allo view is the default foreground view; we only show ego view
            # in foreground if it's the only thing available
            view_dict['ego'] = self.viewer.render(return_rgb_array=True,
                                                  update_foreground=is_human
                                                  and not self.allo_view)

        return view_dict

    def close(self):
        if self.viewer:
            self.viewer.close()
            self.viewer = None

    def debug_print_entity_spec(self):
        """Hacky function to extract specifications for locations & types of
        entities, etc., from the current state of the environment. Good for
        constructing new demo configs from world states."""
        robot_pose = None
        block_colours = []
        block_poses = []
        block_shapes = []
        goal_region_xyhws = []
        goal_region_colours = []
        for entity in self._entities:
            if isinstance(entity, en.Robot):
                _, sig = entity.reconstruct_signature()
                robot_pose = (sig['init_pos'], sig['init_angle'])
            elif isinstance(entity, en.Shape):
                _, sig = entity.reconstruct_signature()
                block_colours.append(sig['colour_name'])
                block_poses.append((sig['init_pos'], sig['init_angle']))
                block_shapes.append(sig['shape_type'])
            elif isinstance(entity, en.GoalRegion):
                _, sig = entity.reconstruct_signature()
                goal_region_xyhws.append(
                    (sig['x'], sig['y'], sig['h'], sig['w']))
                goal_region_colours.append(sig['colour_name'])

        def f_pose(pose):
            # string-format a pose
            (x, y), angle = pose
            return '((%.3f, %.3f), %.3f)' % (x, y, angle)

        def f_xyhw(xyhw):
            return '(%.3f, %.3f, %.3f, %.3f)' % xyhw

        def f_colour(colour):
            # strong-format a colour
            return f'en.ShapeColour.{colour.name.upper()}'

        def f_shape(shape):
            # strong-format a shape
            return f'en.ShapeType.{shape.name.upper()}'

        def lst(f, l):
            # apply function f to each element of l, convert the results to
            # string, then present as string-formatted list
            return '[' + ', '.join(str(f(e)) for e in l) + ']'

        if robot_pose:
            print(f'ROBOT_POSE = {f_pose(robot_pose)}')
        if block_colours:
            print(f'BLOCK_COLOURS = {lst(f_colour, block_colours)}')
            print(f'BLOCK_SHAPES = {lst(f_shape, block_shapes)}')
            print(f'BLOCK_POSES = {lst(f_pose, block_poses)}')
        if goal_region_xyhws:
            print(f'GOAL_REGION_XYHWS = {lst(f_xyhw, goal_region_xyhws)}')
            print(
                f'GOAL_REGION_COLOURS = {lst(f_colour, goal_region_colours)}')



import math
import warnings

from gym.utils import EzPickle
import numpy as np

from magical import geom
from magical.base_env import BaseEnv, ez_init
import magical.entities as en


class MoveToCornerEnv(BaseEnv, EzPickle):
    @ez_init()
    def __init__(self,
                 rand_shape_colour=False,
                 rand_shape_type=False,
                 rand_poses=False,
                 debug_reward=False,
                 **kwargs):
        super().__init__(**kwargs)
        self.rand_shape_colour = rand_shape_colour
        self.rand_shape_type = rand_shape_type
        self.rand_poses = rand_poses
        self.debug_reward = debug_reward
        if self.debug_reward:
            warnings.warn(
                "DEBUG REWARD ENABLED IN MOVE-TO-CORNER ENV! This reward is "
                "ONLY intended for training RL algorithms during debugging, "
                "so don't forget to disable it when benchmarking IL")

    @classmethod
    def make_name(cls, suffix=None):
        base_name = 'MoveToCorner'
        return base_name + (suffix or '') + '-v0'

    def on_reset(self):
        # make the robot
        robot_pos = np.asarray((0.4, -0.0))
        robot_angle = 0.55 * math.pi
        robot = self._make_robot(robot_pos, robot_angle)
        self.add_entities([robot])

        shape_pos = np.asarray((0.1, -0.65))
        shape_angle = 0.13 * math.pi
        shape_colour = 'red'
        shape_type = en.ShapeType.SQUARE
        if self.rand_shape_colour:
            shape_colour = self.rng.choice(
                np.asarray(en.SHAPE_COLOURS, dtype='object'))
        if self.rand_shape_type:
            shape_type = self.rng.choice(
                np.asarray(en.SHAPE_TYPES, dtype='object'))

        shape = self._make_shape(shape_type=shape_type,
                                 colour_name=shape_colour,
                                 init_pos=shape_pos,
                                 init_angle=shape_angle)
        self.add_entities([shape])
        self.__shape_ref = shape

        if self.rand_poses:
            geom.pm_randomise_all_poses(
                self._space, (self._robot, self.__shape_ref),
                self.ARENA_BOUNDS_LRBT,
                self.rng,
                rand_pos=True,
                rand_rot=True,
                rel_pos_linf_limits=self.JITTER_POS_BOUND,
                rel_rot_limits=self.JITTER_ROT_BOUND)

    def score_on_end_of_traj(self):
        robot_pos = np.asarray(self.__shape_ref.shape_body.position)
        # target is top left
        dist = np.linalg.norm(np.asarray([-1.0, 1.0]) - robot_pos)
        succeed_dist = np.sqrt(2) / 2
        furthest_dist = np.sqrt(2)
        drange = (furthest_dist - succeed_dist)
        score = min(1.0, max(0.0, furthest_dist - dist) / drange)
        assert 0 <= score <= 1
        return score

    def step(self, *args, **kwargs):
        obs, rew, done, info = super().step(*args, **kwargs)
        if self.debug_reward:
            # dense reward for training RL
            rew = self.debug_shaped_reward()
        return obs, rew, done, info

    def debug_shaped_reward(self):
        """Compute a heavily shaped reward. This should be sufficient to do RL
        on. It's quite helpful for tuning RL algorithms in new GAIL
        implementations."""
        shape_pos = np.asarray(self.__shape_ref.shape_body.position)
        # shape_pos[0] is meant to be 0, shape_pos[1] is meant to be 1
        target_shape_pos = np.array((0, 1))
        shape_to_corner_dist = np.linalg.norm(shape_pos - target_shape_pos)
        # encourage the robot to get close to the shape, and the shape to get
        # close to the goal
        robot_pos = np.asarray(self._robot.robot_body.position)
        robot_to_shape_dist = np.linalg.norm(robot_pos - shape_pos)
        shaping = -shape_to_corner_dist / 5 \
            - max(robot_to_shape_dist, 0.2) / 20
        return shaping + self.score_on_end_of_traj()



from ctypes import byref

import pyglet
from pyglet import gl
from pyglet.image import ImageException
from pyglet.image.codecs import get_encoders, ImageEncodeException

from magical.pyglet_backport.image.imagebuffer import \
    get_max_color_attachments, Renderbuffer, Framebuffer


class AbstractImage:
    """Abstract class representing an image.
    :Parameters:
        `width` : int
            Width of image
        `height` : int
            Height of image
        `anchor_x` : int
            X coordinate of anchor, relative to left edge of image data
        `anchor_y` : int
            Y coordinate of anchor, relative to bottom edge of image data
    """
    anchor_x = 0
    anchor_y = 0

    def __init__(self, width, height):
        self.width = width
        self.height = height

    def __repr__(self):
        return "{}(size={}x{})".format(self.__class__.__name__, self.width, self.height)

    def get_image_data(self):
        """Get an ImageData view of this image.  
        
        Changes to the returned instance may or may not be reflected in this
        image.
        :rtype: :py:class:`~pyglet.image.ImageData`
        .. versionadded:: 1.1
        """
        raise ImageException('Cannot retrieve image data for %r' % self)

    def get_texture(self, rectangle=False):
        """A :py:class:`~pyglet.image.Texture` view of this image.  
        By default, textures are created with dimensions that are powers of
        two.  Smaller images will return a :py:class:`~pyglet.image.TextureRegion` that covers just
        the image portion of the larger texture.  This restriction is required
        on older video cards, and for compressed textures, or where texture
        repeat modes will be used, or where mipmapping is desired.
        If the `rectangle` parameter is ``True``, this restriction is ignored
        and a texture the size of the image may be created if the driver
        supports the ``GL_ARB_texture_rectangle`` or
        ``GL_NV_texture_rectangle`` extensions.  If the extensions are not
        present, the image already is a texture, or the image has power 2
        dimensions, the `rectangle` parameter is ignored.
        Examine `Texture.target` to determine if the returned texture is a
        rectangle (``GL_TEXTURE_RECTANGLE``) or not (``GL_TEXTURE_2D``).
        Changes to the returned instance may or may not be reflected in this
        image.
        :Parameters:
            `rectangle` : bool
                True if the texture can be created as a rectangle.
                .. versionadded:: 1.1.4.
        :rtype: :py:class:`~pyglet.image.Texture`
        .. versionadded:: 1.1
        """
        raise ImageException('Cannot retrieve texture for %r' % self)

    def get_mipmapped_texture(self):
        """Retrieve a :py:class:`~pyglet.image.Texture` instance with all mipmap levels filled in.
        Requires that image dimensions be powers of 2. 
        :rtype: :py:class:`~pyglet.image.Texture`
        .. versionadded:: 1.1
        """
        raise ImageException('Cannot retrieve mipmapped texture for %r' % self)

    def get_region(self, x, y, width, height):
        """Retrieve a rectangular region of this image.
        :Parameters:
            `x` : int
                Left edge of region.
            `y` : int
                Bottom edge of region.
            `width` : int
                Width of region.
            `height` : int
                Height of region.
        :rtype: AbstractImage
        """
        raise ImageException('Cannot get region for %r' % self)

    def save(self, filename=None, file=None, encoder=None):
        """Save this image to a file.
        :Parameters:
            `filename` : str
                Used to set the image file format, and to open the output file
                if `file` is unspecified.
            `file` : file-like object or None
                File to write image data to.
            `encoder` : ImageEncoder or None
                If unspecified, all encoders matching the filename extension
                are tried.  If all fail, the exception from the first one
                attempted is raised.
        """
        if not file:
            file = open(filename, 'wb')

        if encoder:
            encoder.encode(self, file, filename)
        else:
            first_exception = None
            for encoder in get_encoders(filename):
                try:
                    encoder.encode(self, file, filename)
                    return
                except ImageEncodeException as e:
                    first_exception = first_exception or e
                    file.seek(0)

            if not first_exception:
                raise ImageEncodeException('No image encoders are available')
            raise first_exception

    def blit(self, x, y, z=0):
        """Draw this image to the active framebuffers.
        
        The image will be drawn with the lower-left corner at 
        (``x -`` `anchor_x`, ``y -`` `anchor_y`, ``z``).
        """
        raise ImageException('Cannot blit %r.' % self)

    def blit_into(self, source, x, y, z):
        """Draw `source` on this image.
        `source` will be copied into this image such that its anchor point
        is aligned with the `x` and `y` parameters.  If this image is a 3D
        texture, the `z` coordinate gives the image slice to copy into.
        
        Note that if `source` is larger than this image (or the positioning
        would cause the copy to go out of bounds) then you must pass a
        region of `source` to this method, typically using get_region().
        """
        raise ImageException('Cannot blit images onto %r.' % self)

    def blit_to_texture(self, target, level, x, y, z=0):
        """Draw this image on the currently bound texture at `target`.
        
        This image is copied into the texture such that this image's anchor
        point is aligned with the given `x` and `y` coordinates of the
        destination texture.  If the currently bound texture is a 3D texture,
        the `z` coordinate gives the image slice to blit into.
        """
        raise ImageException('Cannot blit %r to a texture.' % self)


class Texture(AbstractImage):
    """An image loaded into video memory that can be efficiently drawn
    to the framebuffer.
    Typically you will get an instance of Texture by accessing the `texture`
    member of any other AbstractImage.
    :Parameters:
        `region_class` : class (subclass of TextureRegion)
            Class to use when constructing regions of this texture.
        `tex_coords` : tuple
            12-tuple of float, named (u1, v1, r1, u2, v2, r2, ...).  u, v, r
            give the 3D texture coordinates for vertices 1-4.  The vertices
            are specified in the order bottom-left, bottom-right, top-right
            and top-left.
        `target` : int
            The GL texture target (e.g., ``GL_TEXTURE_2D``).
        `level` : int
            The mipmap level of this texture.
    """

    region_class = None  # Set to TextureRegion after it's defined
    tex_coords = (0, 0, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0)
    tex_coords_order = (0, 1, 2, 3)
    level = 0
    images = 1
    x = y = z = 0
    default_min_filter = gl.GL_LINEAR
    default_mag_filter = gl.GL_LINEAR

    def __init__(self, width, height, target, tex_id):
        super(Texture, self).__init__(width, height)
        self.target = target
        self.id = tex_id
        self._context = pyglet.gl.current_context

    def __del__(self):
        try:
            self._context.delete_texture(self.id)
        except:
            pass

    @classmethod
    def create(cls,
               width,
               height,
               target=gl.GL_TEXTURE_2D,
               internalformat=gl.GL_RGBA,
               min_filter=None,
               mag_filter=None):
        """Create a Texture
        Create a Texture with the specified dimentions, target and format.
        On return, the texture will be bound.
        :Parameters:
            `width` : int
                Width of texture in pixels.
            `height` : int
                Height of texture in pixels.
            `target` : int
                GL constant giving texture target to use, typically ``GL_TEXTURE_2D``.
            `internalformat` : int
                GL constant giving internal format of texture; for example, ``GL_RGBA``.
                If ``None``, the texture will be created but not initialized.
            `min_filter` : int
                The minifaction filter used for this texture, commonly ``GL_LINEAR`` or ``GL_NEAREST``
            `mag_filter` : int
                The magnification filter used for this texture, commonly ``GL_LINEAR`` or ``GL_NEAREST``
        :rtype: :py:class:`~pyglet.image.Texture`
        """
        min_filter = min_filter or cls.default_min_filter
        mag_filter = mag_filter or cls.default_mag_filter

        tex_id = gl.GLuint()
        gl.glGenTextures(1, byref(tex_id))
        gl.glBindTexture(target, tex_id.value)
        if target != gl.GL_TEXTURE_2D_MULTISAMPLE:
            gl.glTexParameteri(target, gl.GL_TEXTURE_MIN_FILTER, min_filter)
            gl.glTexParameteri(target, gl.GL_TEXTURE_MAG_FILTER, mag_filter)

        if internalformat is not None:
            blank = (gl.GLubyte * (width * height * 4))()
            gl.glTexImage2D(target, 0, internalformat, width, height, 0,
                            gl.GL_RGBA, gl.GL_UNSIGNED_BYTE, blank)
            gl.glFlush()

        texture = cls(width, height, target, tex_id.value)
        texture.min_filter = min_filter
        texture.mag_filter = mag_filter
        if target is gl.GL_TEXTURE_RECTANGLE:
            texture.tex_coords = (0, 0, 0, width, 0, 0, width, height, 0, 0,
                                  height, 0)
        else:
            texture.tex_coords = cls.tex_coords

        return texture

    def get_image_data(self, z=0, fmt='RGBA', gl_format=gl.GL_RGBA):
        """Get the image data of this texture.
        Changes to the returned instance will not be reflected in this
        texture.
        :Parameters:
            `z` : int
                For 3D textures, the image slice to retrieve.
        :rtype: :py:class:`~pyglet.image.ImageData`
        """
        gl.glBindTexture(self.target, self.id)

        # # Always extract complete RGBA data.  Could check internalformat
        # # to only extract used channels. XXX
        # fmt = 'RGBA'
        # gl_format = gl.GL_RGBA

        gl.glPixelStorei(gl.GL_PACK_ALIGNMENT, 1)
        buffer = (gl.GLubyte *
                  (self.width * self.height * self.images * len(fmt)))()
        gl.glGetTexImage(self.target, self.level, gl_format,
                         gl.GL_UNSIGNED_BYTE, buffer)

        data = pyglet.image.ImageData(self.width, self.height, fmt, buffer)
        if self.images > 1:
            data = data.get_region(0, z * self.height, self.width, self.height)
        return data

    def get_texture(self, rectangle=False):
        if rectangle and not self.target == gl.GL_TEXTURE_RECTANGLE:
            raise gl.ImageException(
                'Texture is not a rectangle, it must be created as a rectangle.'
            )
        return self

    # no implementation of blit_to_texture yet (could use aux buffer)

    def blit(self, x, y, z=0, width=None, height=None):
        x1 = x - self.anchor_x
        y1 = y - self.anchor_y
        x2 = x1 + (width is None and self.width or width)
        y2 = y1 + (height is None and self.height or height)
        vertices = x1, y1, z, x2, y1, z, x2, y2, z, x1, y2, z

        gl.glActiveTexture(gl.GL_TEXTURE0)
        gl.glBindTexture(self.target, self.id)

        pyglet.graphics.draw_indexed(4, gl.GL_TRIANGLES, [0, 1, 2, 0, 2, 3],
                                     ('v3f', vertices),
                                     ('t3f', self.tex_coords))

        gl.glBindTexture(self.target, 0)

    def blit_into(self, source, x, y, z):
        gl.glBindTexture(self.target, self.id)
        source.blit_to_texture(self.target, self.level, x, y, z)

    def get_region(self, x, y, width, height):
        return self.region_class(x, y, 0, width, height, self)

    def get_transform(self, flip_x=False, flip_y=False, rotate=0):
        """Create a copy of this image applying a simple transformation.
        The transformation is applied to the texture coordinates only;
        :py:meth:`~pyglet.image.ImageData.get_image_data` will return the untransformed data.  The
        transformation is applied around the anchor point.
        :Parameters:
            `flip_x` : bool
                If True, the returned image will be flipped horizontally.
            `flip_y` : bool
                If True, the returned image will be flipped vertically.
            `rotate` : int
                Degrees of clockwise rotation of the returned image.  Only 
                90-degree increments are supported.
        :rtype: :py:class:`~pyglet.image.TextureRegion`
        """
        transform = self.get_region(0, 0, self.width, self.height)
        bl, br, tr, tl = 0, 1, 2, 3
        transform.anchor_x = self.anchor_x
        transform.anchor_y = self.anchor_y
        if flip_x:
            bl, br, tl, tr = br, bl, tr, tl
            transform.anchor_x = self.width - self.anchor_x
        if flip_y:
            bl, br, tl, tr = tl, tr, bl, br
            transform.anchor_y = self.height - self.anchor_y
        rotate %= 360
        if rotate < 0:
            rotate += 360
        if rotate == 0:
            pass
        elif rotate == 90:
            bl, br, tr, tl = br, tr, tl, bl
            transform.anchor_x, transform.anchor_y = transform.anchor_y, transform.width - transform.anchor_x
        elif rotate == 180:
            bl, br, tr, tl = tr, tl, bl, br
            transform.anchor_x = transform.width - transform.anchor_x
            transform.anchor_y = transform.height - transform.anchor_y
        elif rotate == 270:
            bl, br, tr, tl = tl, bl, br, tr
            transform.anchor_x, transform.anchor_y = transform.height - transform.anchor_y, transform.anchor_x
        else:
            assert False, 'Only 90 degree rotations are supported.'
        if rotate in (90, 270):
            transform.width, transform.height = transform.height, transform.width
        transform._set_tex_coords_order(bl, br, tr, tl)
        return transform

    def _set_tex_coords_order(self, bl, br, tr, tl):
        tex_coords = (self.tex_coords[:3], self.tex_coords[3:6],
                      self.tex_coords[6:9], self.tex_coords[9:])
        self.tex_coords = tex_coords[bl] + tex_coords[br] + tex_coords[
            tr] + tex_coords[tl]

        order = self.tex_coords_order
        self.tex_coords_order = (order[bl], order[br], order[tr], order[tl])

    def __repr__(self):
        return "{}(id={}, size={}x{})".format(self.__class__.__name__, self.id,
                                              self.width, self.height)



# flake8: noqa
from magical.benchmarks import (AVAILABLE_PREPROCESSORS,
                                DEMO_ENVS_TO_TEST_ENVS_MAP, register_envs)
from magical.reference_demos import try_download_demos  # noqa: F401
from magical.saved_trajectories import (load_demos,
                                        preprocess_demos_with_wrapper,
                                        splice_in_preproc_name)



#!/usr/bin/env python3
"""Convert my old demonstrations so that they use the new action format."""
import gzip
import os
import sys

import click
import cloudpickle
import gym

from magical.baselines.saved_trajectories import (  # this appeases isort
    MAGICALTrajectory, load_demos)
from magical.benchmarks import register_envs

SUFFIX = '.pkl.gz'


@click.command()
@click.option('--out-dir',
              default='demos-converted',
              help='out directory to write to')
@click.argument('target_env_name')
@click.argument('demo_paths', nargs=-1, required=True)
def main(out_dir, target_env_name, demo_paths):
    demo_itr = load_demos(demo_paths)
    if out_dir:
        os.makedirs(out_dir, exist_ok=True)
    register_envs()
    # don't bother seeding env; we'll just hope it's deterministic (which is
    # true of demo variants of magical envs)
    env = gym.make(target_env_name)
    try:
        for old_path, demo_dict in zip(demo_paths, demo_itr):
            print(f"Working on demo at '{old_path}'")
            actions = demo_dict['trajectory'].acts
            observations = [env.reset()]
            rews = []
            infos = []
            for act_num, act in enumerate(actions, 1):
                obs, rew, done, info = env.step(act)
                observations.append(obs)
                rews.append(rew)
                infos.append(info)
                if done:
                    if act_num != len(actions):
                        print(f"Got 'done' at action {act_num}/{len(actions)}")
                    break
            if not done:
                print("Ran out of actions, but env still isn't done (???)")
                print("Going to pad with noops")
                pad_acts = 0
                while not done:
                    obs, rew, done, info = env.step(0)
                    observations.append(obs)
                    rews.append(rew)
                    infos.append(info)
                    pad_acts += 1
                print(f"Padded to {pad_acts} noops")
            score = info.get('eval_score')
            print(f"Traj done. sum(rew) is {sum(rews)}, score is {score}")
            new_traj = MAGICALTrajectory(acts=actions,
                                         obs=observations,
                                         rews=rews,
                                         infos=infos)
            new_dict = {
                'trajectory': new_traj,
                'score': score,
                'env_name': target_env_name,
            }
            old_bn = os.path.basename(old_path)
            new_bn = old_bn.replace(demo_dict['env_name'], target_env_name)
            new_bn = new_bn.replace(' ', '0')  # HACK (e.g. for 'T 9:30:12')
            new_path = os.path.join(out_dir, new_bn)
            print(f"Saving trajectory to '{new_path}'")
            with gzip.GzipFile(new_path, 'wb') as fp:
                cloudpickle.dump(new_dict, fp)
    finally:
        env.close()


if __name__ == '__main__':
    try:
        with main.make_context(sys.argv[0], sys.argv[1:]) as ctx:
            result = main.invoke(ctx)
    except click.ClickException as e:
        e.show()
        sys.exit(e.exit_code)
    except click.exceptions.Exit as e:
        sys.exit(e.exit_code)



"""
2D rendering framework from Gym. MIT license notice below.

Copyright (c) 2016 OpenAI (https://openai.com)

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
of the Software, and to permit persons to whom the Software is furnished to do
so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from __future__ import division
import os
import six
import warnings
import sys

if "Apple" in sys.version:
    if 'DYLD_FALLBACK_LIBRARY_PATH' in os.environ:
        os.environ['DYLD_FALLBACK_LIBRARY_PATH'] += ':/usr/lib'
        # (JDS 2016/04/15): avoid bug on Anaconda 2.3.0 / Yosemite

try:
    import pyglet
except ImportError:
    raise ImportError('''
    Cannot import pyglet. HINT: you can install pyglet directly via 'pip
    install pyglet'. But if you really just want to install all Gym
    dependencies and not have to think about it, 'pip install -e .[all]' or
    'pip install gym[all]' will do it.
    ''')

try:
    from pyglet import gl
except ImportError:
    raise ImportError('''
    Error occurred while running `from pyglet.gl import *` HINT: make sure you
    have OpenGL install. On Ubuntu, you can run 'apt-get install
    python-opengl'. If you're running on a server, you may need a virtual frame
    buffer; something like this should work: 'xvfb-run -s \"-screen 0
    1400x900x24\" python <your_script.py>'
    ''')

import math  # noqa: E402
import numpy as np  # noqa: E402

from magical.pyglet_backport.image import Framebuffer, \
    Renderbuffer, Texture  # noqa: E402

RAD2DEG = 57.29577951308232


def get_display(spec):
    """Convert a display specification (such as :0) into an actual Display
    object.
    Pyglet only supports multiple Displays on Linux.
    """
    if spec is None:
        return None
    elif isinstance(spec, six.string_types):
        return pyglet.canvas.Display(spec)
    else:
        raise RuntimeError(
            'Invalid display spec: {}. (Must be string like :0 or None.)'.
            format(spec))


def get_offscreen_fbo(width, height, msaa_samples=4):
    fbo = Framebuffer()
    # using None for the format specifier in Texture.create means we have to
    # allocate memory ourselves, which is important here because we seem to
    # need to pass nullptr to allocation routine's destination arg (why?).
    if msaa_samples > 1:
        fbo._colour_texture = Texture.create(
            width,
            height,
            target=gl.GL_TEXTURE_2D_MULTISAMPLE,
            internalformat=None)
        gl.glTexImage2DMultisample(gl.GL_TEXTURE_2D_MULTISAMPLE, msaa_samples,
                                   gl.GL_RGB, width, height, True)
    else:
        fbo._colour_texture = Texture.create(width,
                                             height,
                                             internalformat=None)
        gl.glTexImage2D(gl.GL_TEXTURE_2D, 0, gl.GL_RGB, width, height, 0,
                        gl.GL_RGB, gl.GL_UNSIGNED_BYTE, None)
    fbo.attach_texture(gl.GL_FRAMEBUFFER, gl.GL_COLOR_ATTACHMENT0,
                       fbo._colour_texture)
    fbo._depth_rb = Renderbuffer(width,
                                 height,
                                 gl.GL_DEPTH_COMPONENT,
                                 samples=msaa_samples)
    fbo.attach_renderbuffer(gl.GL_FRAMEBUFFER, gl.GL_DEPTH_ATTACHMENT,
                            fbo._depth_rb)
    assert fbo.is_complete, \
        "FramebufferObject not complete after attaching all buffers (bug?); " \
        f"status {{fbo.get_status()}}"
    return fbo


def blit_fbo(width, height, src_id, target_id, target_image=gl.GL_BACK):
    # For drawing a multisampled FBO to a non-multisampled FBO or to the
    # screen. See
    # https://www.khronos.org/opengl/wiki/Multisampling#Allocating_a_Multisample_Render_Target
    gl.glBindFramebuffer(gl.GL_READ_FRAMEBUFFER, src_id)
    gl.glBindFramebuffer(gl.GL_DRAW_FRAMEBUFFER, target_id)
    gl.glDrawBuffer(target_image)
    gl.glBlitFramebuffer(0, 0, width, height, 0, 0, width, height,
                         gl.GL_COLOR_BUFFER_BIT, gl.GL_NEAREST)


class Viewer(object):
    def __init__(self,
                 width,
                 height,
                 visible,
                 display=None,
                 background_rgb=(1, 1, 1)):
        display = get_display(display)

        self.width = width
        self.height = height
        self.background_rgb = background_rgb
        self.window = pyglet.window.Window(width=width,
                                           height=height,
                                           display=display,
                                           visible=visible)
        if not visible:
            # get around stupid bug (?) where OpenGL refuses to render anything
            # to FBOs until the window is displayed
            self.window.set_visible(True)
            self.window.set_visible(False)
        # need to use a second FBO to actually get image data
        self.render_fbo = get_offscreen_fbo(width, height, msaa_samples=1)
        self.no_msaa_fbo = get_offscreen_fbo(width, height, msaa_samples=1)
        self.window.on_close = self.window_closed_by_user
        self.isopen = True
        self.reset_geoms()
        self.transforms = Transform()

        gl.glEnable(gl.GL_BLEND)
        # tricks from OpenAI's multiagent particle env repo:
        # gl.glEnable(gl.GL_MULTISAMPLE)
        gl.glEnable(gl.GL_LINE_SMOOTH)
        # gl.glHint(gl.GL_LINE_SMOOTH_HINT, gl.GL_DONT_CARE)
        gl.glHint(gl.GL_LINE_SMOOTH_HINT, gl.GL_NICEST)
        gl.glLineWidth(2.0)
        gl.glBlendFunc(gl.GL_SRC_ALPHA, gl.GL_ONE_MINUS_SRC_ALPHA)

    def reset_geoms(self):
        self.geoms = []
        self.onetime_geoms = []

    def close(self):
        self.window.close()

    def window_closed_by_user(self):
        self.isopen = False

    def set_bounds(self, left, right, bottom, top, rotation=0.0):
        assert right > left and top > bottom
        scalex = self.width / (right - left)
        scaley = self.height / (top - bottom)
        self.transform = Transform(translation=(-left * scalex,
                                                -bottom * scaley),
                                   scale=(scalex, scaley))

    def set_cam_follow(self, source_xy_world, target_xy_01, viewport_hw_world,
                       rotation):
        """Set camera so that point at `source_xy_world` (in world coordinates)
        appears at `screen_target_xy` (in screen coordinates, in [0,1] on each
        axis), and so that viewport covers region defined by
        `viewport_hw_world` in world coordinates. Oh, and the world is rotated
        by `rotation` around the point `source_xy_world` before doing anything
        else."""
        world_h, world_w = viewport_hw_world
        scalex = self.width / world_w
        scaley = self.height / world_h
        target_x_01, target_y_01 = target_xy_01
        self.transform = TransformEgocentric(centre=source_xy_world,
                                             newpos=(world_w * target_x_01,
                                                     world_h * target_y_01),
                                             scale=(scalex, scaley),
                                             rotation=rotation)

    def add_geom(self, geom):
        self.geoms.append(geom)

    def add_onetime(self, geom):
        self.onetime_geoms.append(geom)

    def render(self, return_rgb_array=False, update_foreground=True):
        # switch to window and ONLY render to FBO
        self.window.switch_to()
        self.render_fbo.bind()

        # actual rendering
        gl.glClearColor(*self.background_rgb, 1)
        self.window.clear()
        self.window.dispatch_events()
        self.transform.enable()
        for geom in self.geoms:
            geom.render()
        for geom in self.onetime_geoms:
            geom.render()
        self.transform.disable()
        self.onetime_geoms = []

        # done, don't need FBO in main context
        self.render_fbo.unbind()

        # optionally write RGB array
        arr = None
        if return_rgb_array:
            # this initial blit call will crash if you're calling it from a
            # fork()ed subprocess; use the 'spawn' MP method instead of 'fork'
            # to make it work
            blit_fbo(self.width, self.height, self.render_fbo.id,
                     self.no_msaa_fbo.id, gl.GL_COLOR_ATTACHMENT0)
            image_data = self.no_msaa_fbo._colour_texture.get_image_data(
                fmt='RGB', gl_format=gl.GL_RGB)
            arr = np.frombuffer(image_data.get_data(), dtype=np.uint8)
            arr = arr.reshape(self.height, self.width, 3)[::-1]

        # optionally blit to main window (should be on by default, but we can
        # skip it if we only want an offscreen render)
        if update_foreground:
            gl.glClearColor(*self.background_rgb, 1)
            self.window.clear()
            blit_fbo(self.width, self.height, self.render_fbo.id, 0)
            self.window.flip()

        return arr if return_rgb_array else self.isopen

    # Convenience
    def draw_circle(self, radius=10, res=30, filled=True, **attrs):
        geom = make_circle(radius=radius, res=res, filled=filled)
        _add_attrs(geom, attrs)
        self.add_onetime(geom)
        return geom

    def draw_polygon(self, v, filled=True, **attrs):
        geom = make_polygon(v=v, filled=filled)
        _add_attrs(geom, attrs)
        self.add_onetime(geom)
        return geom

    def draw_polyline(self, v, **attrs):
        geom = make_polyline(v=v)
        _add_attrs(geom, attrs)
        self.add_onetime(geom)
        return geom

    def draw_line(self, start, end, **attrs):
        geom = Line(start, end)
        _add_attrs(geom, attrs)
        self.add_onetime(geom)
        return geom

    def get_array(self):
        self.window.flip()
        image_data = pyglet.image.get_buffer_manager() \
            .get_color_buffer().get_image_data()
        self.window.flip()
        arr = np.fromstring(image_data.data, dtype=np.uint8, sep='')
        arr = arr.reshape(self.height, self.width, 4)
        return arr[::-1, :, 0:3]

    def __del__(self):
        try:
            self.close()
        except Exception as ex:
            warnings.warn(f'Exception on env auto-close, you probably need '
                          f'to manually close your Gym envs. Error messsage: '
                          f'{ex!s}')


def _add_attrs(geom, attrs):
    if "color" in attrs:
        geom.set_color(*attrs["color"])
    if "linewidth" in attrs:
        geom.set_linewidth(attrs["linewidth"])


class Geom(object):
    def __init__(self):
        self._color = Color((0, 0, 0, 1.0))
        self.attrs = [self._color]

    def render(self):
        for attr in reversed(self.attrs):
            attr.enable()
        self.render1()
        for attr in self.attrs:
            attr.disable()

    def render1(self):
        raise NotImplementedError

    def add_attr(self, attr):
        self.attrs.append(attr)
        return self

    def set_color(self, r, g, b):
        self._color.vec4 = (r, g, b, 1)
        return self


class Attr(object):
    def enable(self):
        raise NotImplementedError

    def disable(self):
        pass


class Transform(Attr):
    def __init__(self, translation=(0.0, 0.0), rotation=0.0, scale=(1, 1)):
        self.set_translation(*translation)
        self.set_rotation(rotation)
        self.set_scale(*scale)

    def enable(self):
        gl.glPushMatrix()
        # translate to GL loc ppint
        gl.glTranslatef(self.translation[0], self.translation[1], 0)
        gl.glRotatef(RAD2DEG * self.rotation, 0, 0, 1.0)
        gl.glScalef(self.scale[0], self.scale[1], 1)

    def disable(self):
        gl.glPopMatrix()

    def set_translation(self, newx, newy):
        self.translation = (float(newx), float(newy))
        return self

    def set_rotation(self, new):
        self.rotation = float(new)
        return self

    def set_scale(self, newx, newy):
        self.scale = (float(newx), float(newy))
        return self


class TransformEgocentric(Attr):
    """Transform class for egocentric top-down camera. Rotates the world around
    a particular point, moves that point to another position, then scales the
    world."""
    def __init__(self, centre, newpos, rotation, scale):
        self.centre = centre
        self.newpos = newpos
        self.rotation = rotation
        self.scale = scale

    def enable(self):
        gl.glPushMatrix()
        gl.glScalef(self.scale[0], self.scale[1], 1)
        gl.glTranslatef(self.newpos[0], self.newpos[1], 0)
        gl.glRotatef(RAD2DEG * -self.rotation, 0, 0, 1.0)
        gl.glTranslatef(-self.centre[0], -self.centre[1], 0)

    def disable(self):
        gl.glPopMatrix()


class Color(Attr):
    def __init__(self, vec4):
        self.vec4 = vec4

    def enable(self):
        gl.glColor4f(*self.vec4)


class LineStyle(Attr):
    def __init__(self, style):
        self.style = style

    def enable(self):
        gl.glEnable(gl.GL_LINE_STIPPLE)
        gl.glLineStipple(1, self.style)

    def disable(self):
        gl.glDisable(gl.GL_LINE_STIPPLE)


class LineWidth(Attr):
    def __init__(self, stroke):
        self.stroke = stroke

    def enable(self):
        gl.glLineWidth(self.stroke)


class Point(Geom):
    def __init__(self):
        Geom.__init__(self)

    def render1(self):
        gl.glBegin(gl.GL_POINTS)  # draw point
        gl.glVertex3f(0.0, 0.0, 0.0)
        gl.glEnd()


class FilledPolygon(Geom):
    def __init__(self, v):
        Geom.__init__(self)
        self.v = v

    def render1(self):
        if len(self.v) == 4:
            gl.glBegin(gl.GL_QUADS)
        elif len(self.v) > 4:
            gl.glBegin(gl.GL_POLYGON)
        else:
            gl.glBegin(gl.GL_TRIANGLES)
        for p in self.v:
            gl.glVertex3f(p[0], p[1], 0)  # draw each vertex
        gl.glEnd()


def make_circle(radius=10, res=30, filled=True):
    points = []
    for i in range(res):
        ang = 2 * math.pi * i / res
        points.append((math.cos(ang) * radius, math.sin(ang) * radius))
    if filled:
        return FilledPolygon(points)
    else:
        return PolyLine(points, True)


def make_rect(width=10, height=10, filled=True):
    rad_h = height / 2
    rad_w = width / 2
    points = [(-rad_w, rad_h), (rad_w, rad_h), (rad_w, -rad_h),
              (-rad_w, -rad_h)]
    if filled:
        return FilledPolygon(points)
    else:
        return PolyLine(points, True)


def make_square(side_length=10, filled=True):
    return make_rect(side_length, side_length, filled)


def make_polygon(v, filled=True):
    if filled:
        return FilledPolygon(v)
    else:
        return PolyLine(v, True)


def make_polyline(v):
    return PolyLine(v, False)


def make_capsule(length, width):
    l, r, t, b = 0, length, width / 2, -width / 2
    box = make_polygon([(l, b), (l, t), (r, t), (r, b)])
    circ0 = make_circle(width / 2)
    circ1 = make_circle(width / 2)
    circ1.add_attr(Transform(translation=(length, 0)))
    geom = Compound([box, circ0, circ1])
    return geom


class Compound(Geom):
    def __init__(self, gs):
        Geom.__init__(self)
        self.gs = gs

    def render1(self):
        for g in self.gs:
            g.render()


class PolyLine(Geom):
    def __init__(self, v, close):
        Geom.__init__(self)
        self.v = v
        self.close = close
        self.linewidth = LineWidth(1)
        self.add_attr(self.linewidth)

    def render1(self):
        gl.glBegin(gl.GL_LINE_LOOP if self.close else gl.GL_LINE_STRIP)
        for p in self.v:
            gl.glVertex3f(p[0], p[1], 0)  # draw each vertex
        gl.glEnd()

    def set_linewidth(self, x):
        self.linewidth.stroke = x


class Line(Geom):
    def __init__(self, start=(0.0, 0.0), end=(0.0, 0.0)):
        Geom.__init__(self)
        self.start = start
        self.end = end
        self.linewidth = LineWidth(1)
        self.add_attr(self.linewidth)

    def render1(self):
        gl.glBegin(gl.GL_LINES)
        gl.glVertex2f(*self.start)
        gl.glVertex2f(*self.end)
        gl.glEnd()


class Image(Geom):
    def __init__(self, fname, width, height):
        Geom.__init__(self)
        self.width = width
        self.height = height
        img = pyglet.image.load(fname)
        self.img = img
        self.flip = False

    def render1(self):
        self.img.blit(-self.width / 2,
                      -self.height / 2,
                      width=self.width,
                      height=self.height)


# ================================================================


class SimpleImageViewer(object):
    def __init__(self, display=None, maxwidth=500):
        self.window = None
        self.isopen = False
        self.display = display
        self.maxwidth = maxwidth

    def imshow(self, arr):
        if self.window is None:
            height, width, _channels = arr.shape
            if width > self.maxwidth:
                scale = self.maxwidth / width
                width = int(scale * width)
                height = int(scale * height)
            self.window = pyglet.window.Window(width=width,
                                               height=height,
                                               display=self.display,
                                               vsync=False,
                                               resizable=True)
            self.width = width
            self.height = height
            self.isopen = True

            @self.window.event
            def on_resize(width, height):
                self.width = width
                self.height = height

            @self.window.event
            def on_close():
                self.isopen = False

        assert len(arr.shape) == 3, \
            "You passed in an image with the wrong number of dimensions"
        image = pyglet.image.ImageData(arr.shape[1],
                                       arr.shape[0],
                                       'RGB',
                                       arr.tobytes(),
                                       pitch=arr.shape[1] * -3)
        gl.glTexParameteri(gl.GL_TEXTURE_2D, gl.GL_TEXTURE_MAG_FILTER,
                           gl.GL_NEAREST)
        texture = image.get_texture()
        texture.width = self.width
        texture.height = self.height
        self.window.clear()
        self.window.switch_to()
        self.window.dispatch_events()
        texture.blit(0, 0)  # draw
        self.window.flip()

    def close(self):
        if self.isopen and sys.meta_path:
            # ^^^ check sys.meta_path to avoid 'ImportError: sys.meta_path is
            # None, Python is likely shutting down'
            self.window.close()
            self.isopen = False

    def __del__(self):
        self.close()



"""Tools for downloading reference demonstrations."""
import argparse
import datetime
import logging
import os
import tempfile
import zipfile

import requests

__all__ = ['try_download_demos']

REFERENCE_DEMO_ZIP = "https://github.com/qxcv/magical-data/archive/master.zip"
REFERENCE_DEMO_ZIP_ROOT = "magical-data-master"
DEFAULT_LOCATION = "demos"
DONE_FILE = ".download-done"


def try_download_demos(dest=DEFAULT_LOCATION, progress=True):
    """Download demonstrations to 'dest', if they do not already exist.

    Args:
        dest (str): destination directory to write demonstrations to.
        progress (bool): should progress messages be displayed to stdout?"""
    zip_url = REFERENCE_DEMO_ZIP
    done_path = os.path.join(dest, DONE_FILE)

    if os.path.exists(done_path):
        # already downloaded
        logging.info(f"Demonstrations appear to already be in '{dest}'; to "
                     "force download, delete that directory and try again")
        return

    with tempfile.NamedTemporaryFile("wb") as temp_fp:
        # download zipped copy of master
        logging.info(
            f"Destination not found. Downloading data from f'{zip_url}' to a "
            "temporary file")
        mb_downloaded = _download_file(zip_url, temp_fp, progress=progress)
        temp_fp.flush()
        logging.info(f"Download done, got {mb_downloaded:.2f}MB")

        # extract each file in the zipped copy to destination directory
        _recursive_extract(zip_fp=temp_fp.name,
                           dest_dir=dest,
                           member_prefix=REFERENCE_DEMO_ZIP_ROOT)

    # create a file to indicate that we have already downloaded this data
    with open(done_path, "w") as fp:
        print(
            f"Data downloaded from '{zip_url}' on "
            f"{datetime.datetime.now().isoformat()}",
            file=fp)


class DownloadError(Exception):
    pass


def _download_file(url,
                   dest_fp,
                   progress_mb=10,
                   progress=True,
                   bufsize=32 * 1024):
    try:
        # launch request & make sure we get a non-error code
        res = requests.get(url, stream=True)
        res.raise_for_status()

        # TODO: infer content size. It's not as simple as looking at
        # Content-Length, since response seems to have Transfer-Encoding:
        # chunked. Somehow browsers do this anyway, though?

        # download one chunk at a time, logging as we go
        next_report_mb = 0
        bytes_downloaded = 0
        for chunk in res.iter_content(bufsize):
            # read one chunk
            dest_fp.write(chunk)
            bytes_downloaded += len(chunk)

            # logging
            mb_downloaded = bytes_downloaded / float(1024**2)
            if progress and mb_downloaded >= next_report_mb:
                print(f'{mb_downloaded:.2f}MiB downloaded')
                next_report_mb += progress_mb

        return mb_downloaded

    except (IOError, requests.HTTPError) as ex:
        raise DownloadError(f"Error while downloading '{url}': {ex}")


def _recursive_extract(zip_fp, dest_dir, member_prefix=None, overwrite=True):
    """Recursively extract a directory from a zip file to another directory on
    the filesystem, overwriting files as necessary."""

    # we're going to use these to figure out relative paths
    if member_prefix is not None:
        split_zip_path = member_prefix.split(os.path.sep)
        prefix_len = len(split_zip_path)
    else:
        split_zip_path = ()
        prefix_len = 0

    with zipfile.ZipFile(zip_fp, 'r') as zip_fp:
        # now iterate over each member of the zip file, and extract any members
        # that are (1) actually files (not directories), and (2) fall under
        # member_prefix in the zip file
        for zip_rel_path in zip_fp.namelist():
            split_rel_path = zip_rel_path.split(os.path.sep)
            if split_rel_path and split_rel_path[-1] == '':
                # skip directories (that end with a "/")
                continue
            if split_rel_path[:len(split_zip_path)] != split_zip_path:
                # skip directories that don't begin with 'member_prefix'
                continue

            # make the directory that we will write this file to
            rel_parts = os.path.sep.join(split_rel_path[prefix_len:])
            out_path = os.path.join(dest_dir, rel_parts)
            save_dir = os.path.dirname(out_path)
            if save_dir:
                os.makedirs(save_dir, exist_ok=True)

            # write the file
            logging.debug(f"Extracting '{zip_rel_path}' -> '{out_path}'")
            with open(out_path, 'wb') as fs_fp:
                fs_fp.write(zip_fp.read(zip_rel_path))


def _main():
    logging.basicConfig(level=logging.DEBUG)

    parser = argparse.ArgumentParser(
        description="Download MAGICAL reference demonstrations")
    parser.add_argument(
        "--dest",
        help=f"destination to download to (default: '{DEFAULT_LOCATION}')",
        default=DEFAULT_LOCATION,
        type=str)
    args = parser.parse_args()

    try_download_demos(dest=args.dest)


if __name__ == '__main__':
    _main()



#!/usr/bin/env python3
"""Render a list of demonstrations as a long, contiguous video."""

import collections
import os
import sys

import click
import numpy as np
import skvideo.io as vidio

from magical.baselines.saved_trajectories import load_demos


def get_frames(demo):
    frame_dicts = demo['trajectory'].obs
    keys = sorted(frame_dicts[0].keys())
    by_key = collections.defaultdict(list)
    for frame_dict in frame_dicts:
        for key in keys:
            by_key[key].append(frame_dict[key])
    # concatenate along time axis
    by_key = {k: np.stack(l, axis=0) for k, l in by_key.items()}
    # concatenate along width axis to show everything
    frames = np.concatenate([by_key[key] for key in keys], axis=2)
    return frames


@click.command()
@click.option('--out-path', default='demo-video.mp4', help='output video path')
@click.option('--fps', default=32, help='frame rate for rendering')
@click.argument('demo_paths', nargs=-1, required=True)
def main(out_path, fps, demo_paths):
    all_demos = load_demos(demo_paths)
    frame_segments = map(get_frames, all_demos)
    out_dir = os.path.dirname(out_path)
    if out_dir:
        print(f"Will make dir '{out_dir}' if it does not exist yet")
        os.makedirs(out_dir, exist_ok=True)
    print(f"Writing video to '{out_path}'")
    writer = vidio.FFmpegWriter(out_path, outputdict={
                     '-r': str(fps),
                     '-vcodec': 'libx264',
                     '-pix_fmt': 'yuv420p',
                 })
    nframes = 0
    for frame_segment in frame_segments:
        print(f"Writing frame segment ({nframes} frames written so far)")
        for frame in frame_segment:
            writer.writeFrame(frame)
            nframes += 1
    print(f"Done! ({nframes} frames written in total)")


if __name__ == '__main__':
    try:
        with main.make_context(sys.argv[0], sys.argv[1:]) as ctx:
            result = main.invoke(ctx)
    except click.ClickException as e:
        e.show()
        sys.exit(e.exit_code)
    except click.exceptions.Exit as e:
        sys.exit(e.exit_code)



import math

from gym.utils import EzPickle
import numpy as np

from magical.base_env import BaseEnv, ez_init
import magical.entities as en
import magical.geom as geom


class MatchRegionsEnv(BaseEnv, EzPickle):
    """Need to push blocks of a certain colour to the corresponding coloured
    region. Aim is for the robot to generalise _that_ rule instead of
    generalising others (e.g. "always move squares to this position" or
    something)."""

    @ez_init()
    def __init__(
            self,
            rand_target_colour=False,
            # shape_randomisation=ShapeRandLevel.NONE,
            rand_shape_type=False,
            rand_shape_count=False,
            rand_layout_minor=False,
            rand_layout_full=False,
            **kwargs):
        super().__init__(**kwargs)
        self.rand_target_colour = rand_target_colour
        # self.shape_randomisation = shape_randomisation
        self.rand_shape_type = rand_shape_type
        self.rand_shape_count = rand_shape_count
        self.rand_layout_minor = rand_layout_minor
        self.rand_layout_full = rand_layout_full
        if self.rand_shape_count:
            assert self.rand_layout_full, \
                "if shape count is randomised then layout must also be " \
                "fully randomised"
            assert self.rand_shape_type, \
                "if shape count is randomised then shape type must also be " \
                "randomised"
            assert self.rand_target_colour, \
                "if shape count is randomised then shape colour must also " \
                "be randomised"

    @classmethod
    def make_name(cls, suffix=None):
        base_name = 'MatchRegions'
        return base_name + (suffix or '') + '-v0'

    def on_reset(self):
        # make the robot
        robot_pos = np.asarray((-0.5, 0.1))
        robot_angle = -math.pi * 1.2
        # if necessary, robot pose is randomised below
        robot = self._make_robot(robot_pos, robot_angle)

        # set up target colour/region/pose
        if self.rand_target_colour:
            target_colour = self.rng.choice(en.SHAPE_COLOURS)
        else:
            target_colour = en.ShapeColour.GREEN
        distractor_colours = [
            c for c in en.SHAPE_COLOURS if c != target_colour
        ]
        target_h = 0.7
        target_w = 0.6
        target_x = 0.1
        target_y = 0.7
        if self.rand_layout_minor or self.rand_layout_full:
            if self.rand_layout_minor:
                hw_bound = self.JITTER_TARGET_BOUND
            else:
                hw_bound = None
            target_h, target_w = geom.randomise_hw(
                self.RAND_GOAL_MIN_SIZE,
                self.RAND_GOAL_MAX_SIZE,
                self.rng,
                current_hw=(target_h, target_w),
                linf_bound=hw_bound)
        sensor = en.GoalRegion(target_x, target_y, target_h, target_w,
                               target_colour)
        self.add_entities([sensor])
        self.__sensor_ref = sensor

        # set up spec for remaining blocks
        default_target_types = [
            en.ShapeType.STAR,
            en.ShapeType.SQUARE,
        ]
        default_distractor_types = [
            [],
            [en.ShapeType.PENTAGON],
            [en.ShapeType.CIRCLE, en.ShapeType.PENTAGON],
        ]
        default_target_poses = [
            # (x, y, theta)
            (0.8, -0.7, 2.37),
            (-0.68, 0.72, 1.28),
        ]
        default_distractor_poses = [
            # (x, y, theta)
            [],
            [(-0.05, -0.2, -1.09)],
            [(-0.75, -0.55, 2.78), (0.3, -0.82, -1.15)],
        ]

        if self.rand_shape_count:
            target_count = self.rng.randint(1, 2 + 1)
            distractor_counts = [
                self.rng.randint(0, 2 + 1) for c in distractor_colours
            ]
        else:
            target_count = len(default_target_types)
            distractor_counts = [len(l) for l in default_distractor_types]

        if self.rand_shape_type:
            shape_types_np = np.asarray(en.SHAPE_TYPES, dtype='object')
            target_types = [
                self.rng.choice(shape_types_np) for _ in range(target_count)
            ]
            distractor_types = [[
                self.rng.choice(shape_types_np) for _ in range(dist_count)
            ] for dist_count in distractor_counts]
        else:
            target_types = default_target_types
            distractor_types = default_distractor_types

        if self.rand_layout_full:
            # will do post-hoc randomisation at the end
            target_poses = [(0, 0, 0)] * target_count
            distractor_poses = [[(0, 0, 0)] * dcount
                                for dcount in distractor_counts]
        else:
            target_poses = default_target_poses
            distractor_poses = default_distractor_poses

        assert len(target_types) == target_count
        assert len(target_poses) == target_count
        assert len(distractor_types) == len(distractor_counts)
        assert len(distractor_types) == len(distractor_colours)
        assert len(distractor_types) == len(distractor_poses)
        assert all(
            len(types) == dcount
            for types, dcount in zip(distractor_types, distractor_counts))
        assert all(
            len(poses) == dcount
            for poses, dcount in zip(distractor_poses, distractor_counts))

        self.__target_shapes = [
            self._make_shape(
                shape_type=shape_type,
                colour_name=target_colour,
                init_pos=(shape_x, shape_y),
                init_angle=shape_angle)
            for shape_type, (shape_x, shape_y, shape_angle
                             ) in zip(target_types, target_poses)
        ]
        self.__distractor_shapes = []
        for dist_colour, dist_types, dist_poses \
                in zip(distractor_colours, distractor_types, distractor_poses):
            for shape_type, (shape_x, shape_y, shape_angle) \
                    in zip(dist_types, dist_poses):
                dist_shape = self._make_shape(
                    shape_type=shape_type,
                    colour_name=dist_colour,
                    init_pos=(shape_x, shape_y),
                    init_angle=shape_angle)
                self.__distractor_shapes.append(dist_shape)
        shape_ents = self.__target_shapes + self.__distractor_shapes
        self.add_entities(shape_ents)

        # add this last so it shows up on top, but before layout randomisation,
        # since it needs to be added to the space before randomising
        self.add_entities([robot])

        if self.rand_layout_minor or self.rand_layout_full:
            all_ents = (sensor, robot, *shape_ents)
            if self.rand_layout_minor:
                # limit amount by which position and rotation can be randomised
                pos_limits = self.JITTER_POS_BOUND
                rot_limits = self.JITTER_ROT_BOUND
            else:
                # no limits, can randomise as much as needed
                assert self.rand_layout_full
                pos_limits = rot_limits = None
            # randomise rotations of all entities but goal region
            rand_rot = [False] + [True] * (len(all_ents) - 1)

            geom.pm_randomise_all_poses(
                self._space,
                all_ents,
                self.ARENA_BOUNDS_LRBT,
                rng=self.rng,
                rand_pos=True,
                rand_rot=rand_rot,
                rel_pos_linf_limits=pos_limits,
                rel_rot_limits=rot_limits)

        # set up index for lookups
        self.__ent_index = en.EntityIndex(shape_ents)

    def score_on_end_of_traj(self):
        overlap_ents = self.__sensor_ref.get_overlapping_ents(
            com_overlap=True, ent_index=self.__ent_index)
        target_set = set(self.__target_shapes)
        distractor_set = set(self.__distractor_shapes)
        n_overlap_targets = len(target_set & overlap_ents)
        n_overlap_distractors = len(distractor_set & overlap_ents)
        # what fraction of targets are in the overlap set?
        target_frac_done = n_overlap_targets / len(target_set)
        if len(overlap_ents) == 0:
            contamination_rate = 0
        else:
            # what fraction of the overlap set are distractors?
            contamination_rate = n_overlap_distractors / len(overlap_ents)
        # score guide:
        # - 1 if all target shapes and no distractors are in the target region
        # - 0 if no target shapes in the target region
        # - somewhere in between if some are all target shapes are there, but
        #   there's also contamination (more contamination = worse, fewer
        #   target shapes in target region = worse).
        return target_frac_done * (1 - contamination_rate)



#!/usr/bin/env python3

import cProfile
import datetime

import click
import gym

import magical


def do_eval(env, ntraj):
    for _ in range(ntraj):
        env.reset()
        done = False
        while not done:
            fake_action = env.action_space.sample()
            _, _, done, _ = env.step(fake_action)


@click.command()
@click.option("--ntraj",
              default=100,
              help="number of trajectories to roll out")
@click.option("--seed", default=42, help="env seed")
@click.argument('env_name')
def main(ntraj, env_name, seed):
    """Very simple script to benchmark performance of a particular
    environment. Will simply run a ~hundred trajectories or so and record
    performance."""
    magical.register_envs()
    env = gym.make(env_name)
    env.seed(seed)
    env.action_space.seed(seed)

    dtime = datetime.datetime.now().strftime("%Y-%m-%dT%H:%M:%S")
    out_filename = f'profile-{env_name.lower()}-{dtime}.cprofile'
    print(f"Will write profile to '{out_filename}'")

    cProfile.runctx('do_eval(env, ntraj)',
                    globals(),
                    locals(),
                    filename=out_filename)

    print("Done")


if __name__ == '__main__':
    main()



"""Tools for saving and loading trajectories without requiring the `imitation`
or `tensorflow` packages to be installed."""
import gzip
from pickle import Unpickler
from typing import List, NamedTuple, Optional

import gym
import numpy as np

from magical.benchmarks import DEFAULT_PREPROC_ENTRY_POINT_WRAPPERS


class MAGICALTrajectory(NamedTuple):
    """Trajectory representation compatible with imitation's trajectory data
    class."""

    acts: np.ndarray
    obs: dict
    rews: np.ndarray
    infos: Optional[List[dict]]


class _TrajRewriteUnpickler(Unpickler):
    """Custom unpickler that replaces references to `Trajectory` class in
    `imitation` with custom trajectory class in this module."""
    def find_class(self, module, name):
        # print('find_class(%r, %r)' % (module, name))
        if (module, name) == ('imitation.util.rollout', 'Trajectory') \
          or (module, name) == ('milbench.baselines.saved_trajectories',
                                'MILBenchTrajectory'):
            return MAGICALTrajectory
        return super().find_class(module, name)


def load_demos(demo_paths, rewrite_traj_cls=True, verbose=False):
    """Use GzipFile & pickle to generate a sequence of demo dictionaries from a
    sequence of file paths."""
    n_demos = len(demo_paths)
    for d_num, d_path in enumerate(demo_paths, start=1):
        if verbose:
            print(f"Loading '{d_path}' ({d_num}/{n_demos})")
        with gzip.GzipFile(d_path, 'rb') as fp:
            if rewrite_traj_cls:
                unpickler = _TrajRewriteUnpickler(fp)
            else:
                unpickler = Unpickler(fp)
            this_dict = unpickler.load()
        yield this_dict


def splice_in_preproc_name(base_env_name, preproc_name):
    """Splice the name of a preprocessor into a magical benchmark name. e.g.
    you might start with "MoveToCorner-Demo-v0" and insert "LoResStack" to end
    up with "MoveToCorner-Demo-LoResStack-v0". Will do a sanity check to ensure
    that the preprocessor actually exists."""
    prefix, version = base_env_name.rsplit('-', 1)
    assert preproc_name in DEFAULT_PREPROC_ENTRY_POINT_WRAPPERS, \
        f"no preprocessor named '{preproc_name}', options are " \
        f"{', '.join(DEFAULT_PREPROC_ENTRY_POINT_WRAPPERS)}"
    env_name = f'{prefix}-{preproc_name}-{version}'
    return env_name


class _MockDemoEnv(gym.Wrapper):
    """Mock Gym environment that just returns an observation"""
    def __init__(self, orig_env, trajectory):
        super().__init__(orig_env)
        self._idx = 0
        self._traj = trajectory
        self._traj_length = len(self._traj.acts)

    def reset(self):
        self._idx = 0
        return self._traj.obs[self._idx]

    def step(self, action):
        rew = self._traj.rews[self._idx]
        info = self._traj.infos[self._idx] or {}
        info['_mock_demo_act'] = self._traj.acts[self._idx]
        self._idx += 1
        # ignore action, return next obs
        obs = self._traj.obs[self._idx]
        # it's okay if we run one over the end
        done = self._idx >= self._traj_length
        return obs, rew, done, info


def preprocess_demos_with_wrapper(trajectories,
                                  orig_env_name,
                                  preproc_name=None,
                                  wrapper=None):
    """Preprocess trajectories using one of the built-in environment
    preprocessing pipelines.

    Args:
        trajectories ([Trajectory]): list of trajectories to process.
        orig_env_name (str): name of original environment where trajectories
            were collected. This function will instantiate a temporary instance
            of that environment to get access to an observation space and other
            metadata.
        preproc_name (str or None): name of preprocessor to apply. Should be
            available in
            `magical.benchmarks.DEFAULT_PREPROC_ENTRY_POINT_WRAPPERS`.
        wrapper (callable or None): wrapper constructor. Should take a
            constructed Gym environment and return a wrapped Gym-like
            environment. Either `preproc_name` or `wrapper` must be specified,
            but both cannot be specified at once.

    Returns:
        rv_trajectories ([Trajectory]): equivalent list of trajectories that
            have each been preprocessed with the given wrapper."""
    if preproc_name is not None:
        assert wrapper is None
        wrapper = DEFAULT_PREPROC_ENTRY_POINT_WRAPPERS[preproc_name]
    else:
        assert wrapper is not None
    orig_env = gym.make(orig_env_name)
    wrapped_constructor = wrapper(_MockDemoEnv)
    rv_trajectories = []
    for traj in trajectories:
        mock_env = wrapped_constructor(orig_env=orig_env, trajectory=traj)
        obs = mock_env.reset()
        values = {
            'obs': [],
            'acts': [],
            'rews': [],
            'infos': [],
        }
        values['obs'].append(obs)
        done = False
        while not done:
            obs, rew, done, info = mock_env.step(None)
            acts = info['_mock_demo_act']
            del info['_mock_demo_act']
            values['obs'].append(obs)
            values['acts'].append(acts)
            values['rews'].append(rew)
            values['infos'].append(info)
        # turn obs, acts, and rews into numpy arrays
        stack_values = {
            k: np.stack(vs, axis=0)
            for k, vs in values.items() if k in ['obs', 'acts', 'rews']
        }
        # keep infos as a list (hard to get at elements otherwise)
        stack_values['infos'] = values.get('infos')
        # use type(traj) to preserve either MAGICAL trajectory type or custom
        # type
        new_traj = type(traj)(**stack_values)
        rv_trajectories.append(new_traj)
    return rv_trajectories



#!/usr/bin/env python3
"""Convert my old demonstrations so that they use the new action format."""
import gzip
import os

import cloudpickle
import numpy as np

from magical.baselines.saved_trajectories import (  # this appeases isort
    MAGICALTrajectory, load_demos)
import magical.entities as en

SOURCE_TREE = "demos/"
DEST_TREE = "demos-new/"
SUFFIX = '.pkl.gz'
AC_FLAGS_UD = [
    en.RobotAction.NONE,
    en.RobotAction.UP,
    en.RobotAction.DOWN,
]
AC_FLAGS_LR = [
    en.RobotAction.NONE,
    en.RobotAction.LEFT,
    en.RobotAction.RIGHT,
]
AC_FLAGS_GRIP = [en.RobotAction.OPEN, en.RobotAction.CLOSE]
AC_FLAGLISTS = [AC_FLAGS_UD, AC_FLAGS_LR, AC_FLAGS_GRIP]


def main():
    # dest_tree is where we save things, source_tree is where we read things
    dest_tree = os.path.abspath(DEST_TREE)
    source_tree = os.path.abspath(SOURCE_TREE)
    os.makedirs(dest_tree, exist_ok=True)

    for dirpath, _, filenames in os.walk(source_tree):
        for filename in filenames:
            if not filename.endswith(SUFFIX):
                continue

            # load old trajectory
            source_path = os.path.join(dirpath, filename)
            demo_dict, = load_demos([source_path])
            demo_traj = demo_dict['trajectory']

            # convert actions
            old_acts = demo_traj.acts
            new_acts_list = []
            for act in old_acts:
                assert len(act) == len(AC_FLAGLISTS)
                flags = tuple(flaglist[flag_num]
                              for flag_num, flaglist in zip(act, AC_FLAGLISTS))
                action_id = en.FLAGS_TO_ACTION_ID[flags]
                new_acts_list.append(action_id)
            new_act_tensor = np.asarray(new_acts_list, dtype=np.int32)

            # splice the new actions into a new demonstration with the right
            # trajectory
            new_demo_traj = MAGICALTrajectory(acts=new_act_tensor,
                                              obs=demo_traj.obs,
                                              rews=demo_traj.rews,
                                              infos=demo_traj.infos)
            new_demo_dict = {k: v for k, v in demo_dict.items()}
            new_demo_dict['trajectory'] = new_demo_traj

            # save it in the appropriate place
            prefix_dir = os.path.relpath(dirpath, start=source_tree)
            dest_dir = os.path.join(dest_tree, prefix_dir)
            os.makedirs(dest_dir, exist_ok=True)
            dest_path = os.path.join(dest_dir, filename)
            print(f"'{source_path}' -> '{dest_path}'")
            with gzip.GzipFile(dest_path, 'wb') as fp:
                cloudpickle.dump(new_demo_dict, fp)


if __name__ == '__main__':
    main()
